import logging as _logging
from typing import Optional
import arcgis
from datetime import datetime
from arcgis.features import FeatureSet
from arcgis.gis import GIS
from arcgis.mapping import MapImageLayer
from arcgis.geoprocessing import DataFile, LinearUnit, RasterData
from arcgis.geoprocessing import import_toolbox
from arcgis._impl.common._utils import _validate_url
from ._routing_utils import _create_toolbox

_log = _logging.getLogger(__name__)

_use_async = True

default_origins = {
    "fields": [
        {"alias": "OBJECTID", "name": "OBJECTID", "type": "esriFieldTypeOID"},
        {"alias": "Name", "name": "Name", "type": "esriFieldTypeString", "length": 128},
        {
            "alias": "Target Destination Count",
            "name": "TargetDestinationCount",
            "type": "esriFieldTypeInteger",
        },
        {"alias": "Cutoff", "name": "Cutoff", "type": "esriFieldTypeDouble"},
        {
            "alias": "Curb Approach",
            "name": "CurbApproach",
            "type": "esriFieldTypeSmallInteger",
        },
    ],
    "geometryType": "esriGeometryPoint",
    "displayFieldName": "",
    "exceededTransferLimit": False,
    "spatialReference": {"latestWkid": 4326, "wkid": 4326},
    "features": [],
}

default_destinations = {
    "fields": [
        {"alias": "OBJECTID", "name": "OBJECTID", "type": "esriFieldTypeOID"},
        {"alias": "Name", "name": "Name", "type": "esriFieldTypeString", "length": 128},
        {
            "alias": "Curb Approach",
            "name": "CurbApproach",
            "type": "esriFieldTypeSmallInteger",
        },
    ],
    "geometryType": "esriGeometryPoint",
    "displayFieldName": "",
    "exceededTransferLimit": False,
    "spatialReference": {"latestWkid": 4326, "wkid": 4326},
    "features": [],
}

default_point_barriers = {
    "fields": [
        {"alias": "OBJECTID", "name": "OBJECTID", "type": "esriFieldTypeOID"},
        {"alias": "Name", "name": "Name", "type": "esriFieldTypeString", "length": 128},
        {
            "alias": "Barrier Type",
            "name": "BarrierType",
            "type": "esriFieldTypeInteger",
        },
        {
            "alias": "Additional Time",
            "name": "Additional_Time",
            "type": "esriFieldTypeDouble",
        },
        {
            "alias": "Additional Distance",
            "name": "Additional_Distance",
            "type": "esriFieldTypeDouble",
        },
        {
            "alias": "CurbApproach",
            "name": "CurbApproach",
            "type": "esriFieldTypeSmallInteger",
        },
    ],
    "geometryType": "esriGeometryPoint",
    "displayFieldName": "",
    "exceededTransferLimit": False,
    "spatialReference": {"latestWkid": 4326, "wkid": 4326},
    "features": [],
}

default_line_barriers = {
    "fields": [
        {"alias": "OBJECTID", "name": "OBJECTID", "type": "esriFieldTypeOID"},
        {"alias": "Name", "name": "Name", "type": "esriFieldTypeString", "length": 128},
        {
            "alias": "SHAPE_Length",
            "name": "SHAPE_Length",
            "type": "esriFieldTypeDouble",
        },
    ],
    "geometryType": "esriGeometryPolyline",
    "displayFieldName": "",
    "exceededTransferLimit": False,
    "spatialReference": {"latestWkid": 4326, "wkid": 4326},
    "features": [],
}

default_polygon_barriers = {
    "fields": [
        {"alias": "OBJECTID", "name": "OBJECTID", "type": "esriFieldTypeOID"},
        {"alias": "Name", "name": "Name", "type": "esriFieldTypeString", "length": 128},
        {
            "alias": "Barrier Type",
            "name": "BarrierType",
            "type": "esriFieldTypeInteger",
        },
        {
            "alias": "Scaled Time Factor",
            "name": "ScaledTimeFactor",
            "type": "esriFieldTypeDouble",
        },
        {
            "alias": "Scaled Distance Factor",
            "name": "ScaledDistanceFactor",
            "type": "esriFieldTypeDouble",
        },
        {
            "alias": "SHAPE_Length",
            "name": "SHAPE_Length",
            "type": "esriFieldTypeDouble",
        },
        {"alias": "SHAPE_Area", "name": "SHAPE_Area", "type": "esriFieldTypeDouble"},
    ],
    "geometryType": "esriGeometryPolygon",
    "displayFieldName": "",
    "exceededTransferLimit": False,
    "spatialReference": {"latestWkid": 4326, "wkid": 4326},
    "features": [],
}

default_restrictions = """['Avoid Unpaved Roads', 'Avoid Private Roads', 'Driving an Automobile', 'Through Traffic Prohibited', 'Roads Under Construction Prohibited', 'Avoid Gates', 'Avoid Express Lanes', 'Avoid Carpool Roads']"""

default_attributes = {
    "fields": [
        {"alias": "ObjectID", "name": "OBJECTID", "type": "esriFieldTypeOID"},
        {
            "alias": "AttributeName",
            "name": "AttributeName",
            "type": "esriFieldTypeString",
            "length": 255,
        },
        {
            "alias": "ParameterName",
            "name": "ParameterName",
            "type": "esriFieldTypeString",
            "length": 255,
        },
        {
            "alias": "ParameterValue",
            "name": "ParameterValue",
            "type": "esriFieldTypeString",
            "length": 25,
        },
    ],
    "features": [],
    "displayFieldName": "",
    "exceededTransferLimit": False,
}


def generate_origin_destination_cost_matrix(
    origins: FeatureSet,
    destinations: FeatureSet,
    travel_mode: str = "Custom",
    time_units: str = "Minutes",
    distance_units: str = "Kilometers",
    analysis_region: Optional[str] = None,
    number_of_destinations_to_find: Optional[int] = None,
    cutoff: Optional[float] = None,
    time_of_day: Optional[datetime] = None,
    time_zone_for_time_of_day: str = "Geographically Local",
    point_barriers: Optional[FeatureSet] = None,
    line_barriers: Optional[FeatureSet] = None,
    polygon_barriers: Optional[FeatureSet] = None,
    uturn_at_junctions: str = "Allowed Only at Intersections and Dead Ends",
    use_hierarchy: bool = True,
    restrictions: Optional[str] = None,
    attribute_parameter_values: Optional[FeatureSet] = None,
    impedance: str = "Drive Time",
    origin_destination_line_shape: str = "None",
    save_output_network_analysis_layer: bool = False,
    overrides: Optional[dict] = None,
    time_impedance: Optional[str] = None,
    distance_impedance: Optional[str] = None,
    output_format: Optional[str] = None,
    gis: Optional[GIS] = None,
    future: bool = False,
):
    """

    The ``generate_origin_destination_cost_matrix`` tool creates an origin-destination (OD) cost matrix from multiple origins to multiple destinations.
    An OD cost matrix is a table that contains the travel time and travel distance from each origin to each destination. Additionally, it ranks the
    destinations that each origin connects to in ascending order based on the minimum time or distance required to travel from that origin to each
    destination. The best path on the street network is discovered for each origin-destination pair, and the travel times and travel distances are
    stored as attributes of the output lines. Even though the lines are straight for performance reasons, they always store the travel time and travel
    distance along the street network, not straight-line distance.

    ======================================  ==========================================================================================================================================
    **Parameter**                            **Description**
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    origins                                 Required :class:`~arcgis.features.FeatureSet` . Specify locations that function as starting points in generating the paths to destinations.
                                            You can add up to 200 origins.
                                            When specifying the origins, you can set properties for each one, such as its name or the number of destinations
                                            to find from the origin, by using attributes. The origins can be specified with the following attributes:

                                            * ``Name`` - The name of the origin. The name can be an unique identifier for the origin. The name is included in the output lines (as the OriginName field) and in the output origins (as the Name field) and can be used to join additional information from the tool outputs to the attributes of your origins. If the name is not specified, a unique name prefixed with Location is automatically generated in the output origins. An auto-generated origin name is not included in the output lines.
                                            * ``TargetDestinationCount``-The maximum number of
                                              destinations that must be found for the origin. If a value is not specified, the value from the Number of Destinations
                                              to Find parameter is used. Cutoff-Specify the travel time or travel distance value at which to stop searching for
                                              destinations from the origin. Any destination beyond the cutoff value will not be considered.  The value needs to be
                                              in the units specified by the Time Units parameter if the impedance attribute in your travel mode is time based or in
                                              the units specified by the Distance Units parameter if the impedance attribute in your travel mode is distance based.
                                              If a value is not specified, the value from the Cutoff parameter is used.
                                            * ``CurbApproach`` - Specifies the direction a vehicle may depart from the origin. The field value is specified as one of the
                                              following integers (use the numeric code, not the name in parentheses):

                                              * 0 (Either side of vehicle)-The vehicle can depart the origin in either direction, so a U-turn is allowed at the origin. This setting can be chosen if it is possible and practical for your vehicle to turn around at the origin. This decision may depend on the width of the road and the amount of traffic or whether the origin has a parking lot where vehicles can enter and turn around.
                                              * 1 ( Right side of vehicle)-When the vehicle departs the origin, the origin must be on the right side of the vehicle. A U-turn is prohibited. This is typically used for vehicles such as buses that must depart from the bus stop on the right-hand side.
                                              * 2 (Left side of vehicle)-When the vehicle departs the origin, the curb must be on the left side of the vehicle. A U-turn is prohibited. This is typically used for vehicles such as buses that must depart from the bus stop on the left-hand side.
                                              * 3 (No U-Turn)-For this tool, the No U-turn (3) value functions the same as Either side of vehicle. The CurbApproach property is designed to work with both kinds of national driving standards: right-hand traffic (United States) and left-hand traffic (United Kingdom). First, consider an origin on the left side of a vehicle. It is always on the left side regardless of whether the vehicle travels on the left or right half of the road. What may change with national driving standards is your decision to depart the origin from one of two directions, that is, so it ends up on the right or left side of the vehicle. For example, if you want to depart from an origin and not have a lane of traffic between the vehicle and the origin, you would choose Right side of vehicle (1) in the United States but Left side of vehicle (2) in the United Kingdom.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    destinations                            Required :class:`~arcgis.features.FeatureSet` . Specify locations that function as ending points in generating the paths from origins.
                                            You can add up to 200 destinations. When specifying the destinations, you can set properties for each one, such as its name,
                                            by using attributes. The destinations can be specified with the following attributes:

                                            * ``Name`` - The name of the destination. The name can be an unique identifier for the destination. The name is included in
                                              the output lines (as the DestinationName field) and in the output destinations (as the Name field) and can be used to join
                                              additional information from the tool outputs to the attributes of your destinations.
                                              If the name is not specified, a unique name prefixed with Location is automatically generated in the output destinations.
                                              An auto-generated destination name is not included in the output lines.
                                            * ``CurbApproach`` - Specifies the direction a vehicle may arrive at the destination. The field value is specified as one of the following
                                              integers (use the numeric code, not the name in parentheses):

                                              * 0 (Either side of vehicle)- The vehicle can arrive the destination in either direction, so a U-turn is allowed at the destination. This setting can be chosen if it is possible and practical for your vehicle to turn around at the destination. This decision may depend on the width of the road and the amount of traffic or whether the destination has a parking lot where vehicles can enter and turn around.
                                              * 1 ( Right side of vehicle)- When the vehicle arrives at the destination, the destination must be on the right side
                                                of the vehicle. A U-turn is prohibited. This is typically used for vehicles such as buses that must arrive at the bus stop on the
                                                right-hand side.
                                              * 2 (Left side of vehicle)-When the vehicle arrives at the destination, the curb must be on the left side of the vehicle.
                                                A U-turn is prohibited. This is typically used for vehicles such as buses that must arrive at the bus stop on the left-hand side.
                                              * 3 (No U-Turn)-For this tool, the No U-turn (3) value functions the same as Either side of vehicle. The CurbApproach property is
                                                designed to work with both kinds of national driving standards: right-hand traffic (United States) and left-hand traffic (United Kingdom).
                                                First, consider a destination on the left side of a vehicle. It is always on the left side regardless of whether the vehicle travels
                                                on the left or right half of the road. What may change with national driving standards is your decision to arrive at the destination
                                                from one of two directions, that is, so it ends up on the right or left side of the vehicle. For example, if you want to arrive at the
                                                destination and not have a lane of traffic between the vehicle and the destination, you would choose Right side of vehicle (1) in the
                                                United States but Left side of vehicle (2) in the United Kingdom.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    travel_mode                             Optional string. Specify the mode of transportation to model in the analysis. Travel modes are managed in ArcGIS Online and can be configured by the administrator of your
                                            organization to better reflect your organization's workflows. You need to specify the name of a travel mode supported by your organization.

                                            To get a list of supported travel mode names, run the GetTravelModes tool from the Utilities toolbox available under the same GIS Server connection
                                            you used to access the tool. The GetTravelModes tool adds a table, Supported Travel Modes, to the application. Any value in the Travel Mode Name field from the
                                            Supported Travel Modes table can be specified as input. You can also specify the value from Travel Mode Settings field as input. This speeds up the tool execution as the
                                            tool does not have to lookup the settings based on the travel mode name.

                                            The default value, Custom, allows you to configure your own travel mode using the custom travel mode parameters (UTurn at Junctions, Use Hierarchy, Restrictions, Attribute Parameter Values,  and Impedance).
                                            The default values of the custom travel mode parameters model travelling by car. You may want to choose Custom and set the custom travel mode parameters listed above to model a pedestrian with a fast walking speed
                                            or a truck with a given height, weight, and cargo of certain hazardous materials. You may choose to do this to try out different settings to get desired analysis results.
                                            Once you have identified the analysis settings, you should work with your organization's administrator and save these settings as part of new or existing travel mode so that
                                            everyone in your organization can rerun the analysis with the same settings.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    time_units                              Optional string. Specify the units that should be used to measure and report the total travel time between each origin-destination pair.

                                            Choice list:['Seconds', 'Minutes', 'Hours', 'Days']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    distance_units                          Optional string. Specify the units that should be used to measure and report the total travel distance between each origin-destination pair.

                                            Choice list:['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles', 'NauticalMiles']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    analysis_region                         Optional string. Specify the region in which to perform the analysis. If a value is not specified for this parameter, the tool
                                            will automatically calculate the region name based on the location
                                            of the input points. Setting the name of the region is recommended to speed up the
                                            tool execution.

                                            Choice list:['NorthAmerica', 'SouthAmerica', 'Europe', 'MiddleEastAndAfrica', 'India', 'SouthAsia', 'SouthEastAsia', 'Thailand', 'Taiwan', 'Japan', 'Oceania', 'Greece', 'Korea']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    number_of_destinations_to_find          Optional integer. Specify the maximum number of destinations to find per origin. If a value for this parameter is
                                            not specified, the output matrix includes travel costs from each origin to every destination. Individual origins
                                            can have their own values (specified as the TargetDestinationCount field) that override the Number of Destinations
                                            to Find parameter value.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    cutoff                                  Optional float. Specify the travel time or travel distance value at which to stop searching for destinations
                                            from a given origin. Any destination beyond the cutoff value will not be considered. Individual origins can have
                                            their own values (specified as the Cutoff field) that override the Cutoff parameter value.

                                            The value needs to be in the units specified by the Time Units parameter if the impedance attribute of your travel mode is time based or
                                            in the units specified by the Distance Units parameter if the impedance attribute of your travel mode is distance based.
                                            If a value is not specified, the tool will not enforce any travel time or travel distance limit when searching for destinations.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    time_of_day                             Optional datetime. Specifies the time and date at which the routes should
                                            begin. If you are modeling the driving  travel mode and specify the current date and time as the value
                                            for this parameter, the tool will use live traffic conditions to
                                            find the best routes and the total travel time will be based
                                            on traffic conditions.

                                            Specifying a time of day results in more accurate
                                            routes and estimations of travel times because the
                                            travel times account for the traffic conditions that are applicable
                                            for that date and time.
                                            The Time Zone for Time of Day parameter specifies whether this time and date refer to UTC or the time zone in which the stop is located. The
                                            tool ignores this parameter when Measurement Units isn't set to a time-based unit.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    time_zone_for_time_of_day               Optional string. Specifies the time zone of the Time of Day parameter.

                                            * ``Geographically Local``: The Time of Day parameter refers to the time zone in which the first stop of a route is located.  If you are generating many
                                              routes that start in multiple times zones,  the start times are staggered in Coordinated Universal Time (UTC). For example, a Time of Day value of
                                              10:00 a.m., 2 January, would mean a start time of  10:00 a.m. Eastern Standard Time (3:00 p.m. UTC) for routes beginning in the Eastern Time Zone
                                              and 10:00 a.m. Central Standard Time (4:00 p.m. UTC) for routes beginning in the Central Time Zone. The start times are offset by one hour in UTC.
                                              The arrive and depart times and dates recorded in the output Stops feature class will refer to the local time zone of the first stop for each route.
                                            * ``UTC``: The Time of Day parameter refers to Coordinated Universal Time (UTC). Choose this option if you want to generate a route for a specific time,
                                              such as now, but aren't certain in which time zone the first stop will be located. If you are generating many routes spanning multiple times zones,
                                              the start times in UTC are simultaneous. For example, a Time of Day value of 10:00 a.m., 2 January, would mean a start time of  5:00 a.m. Eastern Standard
                                              Time (UTC-5:00) for routes beginning in the Eastern Time Zone and 4:00 a.m. Central Standard Time (UTC-6:00) for routes beginning in the Central Time
                                              Zone. Both routes would start at 10:00 a.m. UTC. The arrive and depart times and dates recorded in the output Stops feature class will refer to UTC.

                                            Choice list:['Geographically Local', 'UTC']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    point_barriers                          Optional :class:`~arcgis.features.FeatureSet`  . Specify one or more points to act as temporary restrictions or represent additional time or
                                            distance that may be required to travel on the underlying streets. For example, a point
                                            barrier can be used to represent a fallen tree along a street or time delay spent at a railroad crossing.

                                            The tool imposes a limit of 250 points that can be added as barriers.
                                            When specifying the point barriers, you can set properties for each one, such as its name or barrier type,
                                            by using attributes. The point barriers can be specified with the following attributes:

                                            * ``Name``: The name of the barrier.
                                            * ``BarrierType``: Specifies whether the point barrier restricts travel completely or adds time or distance when it is crossed. The value for this attribute is specified as one of the following integers (use the numeric code, not the name in parentheses):

                                              * 0 (Restriction)-Prohibits travel through the barrier. The barrier
                                                is referred to as a restriction point barrier since it acts as a
                                                restriction.
                                              * 2 (Added Cost)-Traveling through the barrier increases the travel
                                                time or distance by the amount specified in the ``Additional_Time`` or ``Additional_Distance`` field. This barrier type is referred to as an added-cost point barrier.

                                            * ``Additional_Time``: Indicates how much travel time is added when the barrier is traversed. This field is applicable only for added-cost barriers and only if the measurement units are time based. This field value must be greater than or equal to zero, and its units are the same as those specified in the Measurement Units parameter.
                                            * ``Additional_Distance``: Indicates how much distance is added when the barrier is traversed. This field is applicable only for added-cost barriers and only if the measurement units are distance based. The field value must be greater than or equal to zero, and its units are the same as those specified in the Measurement Units parameter.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    line_barriers                           Optional :class:`~arcgis.features.FeatureSet`  . Specify one or more lines that prohibit travel anywhere
                                            the lines intersect the streets. For example, a parade or protest
                                            that blocks traffic across several street segments can be modeled
                                            with a line barrier. A line barrier can also quickly fence off
                                            several roads from being traversed, thereby channeling possible
                                            routes away from undesirable parts of the street
                                            network.

                                            The tool imposes a limit on the number of streets you can
                                            restrict using the Line Barriers parameter. While there is no limit on
                                            the number of lines you can specify as line barriers, the combined
                                            number of streets intersected by all the lines cannot exceed
                                            500.
                                            When specifying the line barriers, you can set a name property for each one by using the following attribute:

                                            * ``Name``: The name of the barrier.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    polygon_barriers                        Optional :class:`~arcgis.features.FeatureSet`  . Specify polygons that either completely restrict travel or
                                            proportionately scale the time or distance required to travel on
                                            the streets intersected by the polygons.

                                            The service imposes a limit on the number of streets you
                                            can restrict using the Polygon Barriers parameter. While there is
                                            no limit on the number of polygons you can specify as the polygon
                                            barriers, the combined number of streets intersected by all the
                                            polygons should not exceed 2,000.
                                            When specifying the polygon barriers, you can set properties for each one, such as its name or barrier type,
                                            by using attributes. The polygon barriers can be specified with the following attributes:

                                            * ``Name``: The name of the barrier.
                                            * ``BarrierType``: Specifies whether the barrier restricts travel completely or scales the time or distance for traveling through it. The field value is specified as one of the following integers (use the numeric code, not the name in parentheses):

                                              * 0 (Restriction)-Prohibits traveling through any part of the barrier. The barrier is referred to as a restriction polygon barrier since it prohibits traveling on streets intersected by the barrier. One use of this type of barrier is to model floods covering areas of the street that make traveling on those streets impossible.
                                              * 1 (Scaled Cost)-Scales the time or distance required to travel the
                                                underlying streets by a factor specified using the ScaledTimeFactor
                                                or ScaledDistanceFactor fields. If the streets are partially
                                                covered by the barrier, the travel time or distance is apportioned
                                                and then scaled. For example, a factor 0.25 would mean that travel
                                                on underlying streets is expected to be four times faster than
                                                normal. A factor of 3.0 would mean it is expected to take three
                                                times longer than normal to travel on underlying streets. This
                                                barrier type is referred to as a scaled-cost polygon barrier. It
                                                might be used to model storms that reduce travel speeds in specific
                                                regions.

                                            * ``ScaledTimeFactor``: This is the factor by which the travel time of the streets
                                              intersected by the barrier is multiplied. This field is applicable
                                              only for scaled-cost barriers and only if the measurement units are time
                                              based. The field value must be greater than zero.
                                            * ``ScaledDistanceFactor``: This is the factor by which the distance of the streets
                                              intersected by the barrier is multiplied. This attribute is
                                              applicable only for scaled-cost barriers and only if the measurement
                                              units are distance based. The attribute value must be greater than
                                              zero.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    uturn_at_junctions                      Optional string. The U-Turn policy at junctions. Allowing U-turns implies the solver can turn around at a
                                            junction and double back on the same street.

                                            Given that junctions represent street intersections and dead ends, different vehicles  may be able to turn around at
                                            some junctions but not at others-it depends on whether the junction represents an intersection or dead end.
                                            To accommodate, the U-turn policy parameter is implicitly specified by how many edges, or streets, connect to the
                                            junction, which is known as junction valency. The acceptable values for this parameter are listed below; each is
                                            followed by a description of its meaning in terms of junction valency.

                                            Choice list:['Allowed', 'Not Allowed', 'Allowed Only at Dead Ends', 'Allowed Only at Intersections and Dead Ends']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    use_hierarchy                           Optional boolean. Specify whether hierarchy should be used when finding the shortest paths between stops.
                                            Checked (True): Use hierarchy when finding routes. When hierarchy is used, the tool prefers higher-order  streets (such as
                                            freeways) to lower-order streets (such as local roads), and can be used
                                            to simulate the driver preference of traveling on freeways instead
                                            of local roads even if that means a longer trip. This is especially
                                            true when finding routes to faraway locations, because drivers on long-distance trips tend to prefer traveling on
                                            freeways where stops, intersections, and turns can be avoided. Using hierarchy is computationally faster,
                                            especially for long-distance routes, since the tool can determine the
                                            best route from a relatively smaller subset of streets.
                                            Unchecked (False):
                                            Do not use hierarchy when finding routes. If
                                            hierarchy is not used, the tool considers all the streets and doesn't
                                            prefer higher-order streets when finding the route. This is often
                                            used when finding short routes within a city.

                                            The tool automatically reverts to using hierarchy if the
                                            straight-line distance between facilities and demand points is
                                            greater than 50 miles (80.46
                                            kilometers), even if you have set this parameter to not use hierarchy.
                                            This parameter is ignored unless Travel Mode is set to Custom. When modeling a custom walking mode,
                                            it is recommended to turn off hierarchy since the hierarchy is designed for motorized vehicles.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    restrictions                            Optional string. Specify which restrictions should be honored by the tool when finding the best routes.
                                            A restriction represents a driving preference or requirement. In most cases, restrictions cause roads to be
                                            prohibited, but they can also cause them to be avoided or preferred. For instance, using an Avoid Toll Roads
                                            restriction will result in a route that will include toll roads only when it is absolutely required to travel
                                            on toll roads in order to visit a stop. Height Restriction makes it possible to route around any clearances
                                            that are lower than the height of your vehicle. If you are carrying corrosive materials on your vehicle,
                                            using the Any Hazmat Prohibited restriction prevents hauling the materials along roads where it is marked
                                            as illegal to do so. The values you provide for this parameter are ignored unless Travel Mode is set to
                                            Custom. Below is a list of available restrictions and a short description.
                                            Some restrictions require an additional value to be specified for their desired use. This value needs to be associated
                                            with the restriction name and a specific parameter intended to work with the restriction. You can identify such restrictions if their
                                            names appear under the AttributeName column in the Attribute
                                            Parameter Values parameter. The ParameterValue field should be
                                            specified in the Attribute Parameter Values parameter for the
                                            restriction to be correctly used when finding traversable roads.
                                            Some restrictions are supported only in certain countries; their availability is stated by region in the list below.
                                            Of the restrictions that have limited availability within a region, you can check whether the restriction is available
                                            in a particular country by looking at the table in the Country List section of the Data coverage for network analysis
                                            services web page. If a country has a value of  Yes in the Logistics Attribute column, the restriction with select
                                            availability in the region is supported in that country. If you specify restriction names that are not available
                                            in the country where your incidents are located, the service ignores the invalid restrictions. The service also
                                            ignores restrictions whose Restriction Usage parameter value is between 0 and 1 (see the Attribute Parameter Value
                                            parameter). It prohibits all restrictions whose Restriction Usage parameter value is greater than 0.

                                            ========================================  ================================================
                                            **Parameter**                             **Description**
                                            ----------------------------------------  ------------------------------------------------
                                            Any Hazmat Prohibited                     The results will not include roads
                                                                                      where transporting any kind of hazardous material is
                                                                                      prohibited.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Carpool Roads                       The results will avoid roads that are
                                                                                      designated exclusively for carpool (high-occupancy)
                                                                                      vehicles.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Express Lanes                       The results will avoid roads designated
                                                                                      as express lanes.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Ferries                             The results will avoid ferries.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Gates                               The results will avoid roads where there are
                                                                                      gates such as keyed access or guard-controlled
                                                                                      entryways.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Limited Access Roads                The results will avoid roads
                                                                                      that are limited access highways.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Private Roads                       The results will avoid roads that are
                                                                                      not publicly owned and maintained.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Roads Unsuitable for Pedestrians    The result will avoid roads that are unsuitable for pedestrians.
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Stairways                           The result will avoid all stairways on a pedestrian suitable route.
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Toll Roads                          The results will avoid toll
                                                                                      roads.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Toll Roads for Trucks               The result will avoid all toll roads for trucks
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Truck Restricted Roads              The result will avoid roads where trucks are not allowed except when making deliveries.
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Unpaved Roads                       The results will avoid roads that are
                                                                                      not paved (for example, dirt, gravel, and so on).
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Axle Count Restriction                    The results will not include roads
                                                                                      where trucks with the specified number of axles are prohibited. The
                                                                                      number of axles can be specified using the Number of Axles
                                                                                      restriction parameter.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Driving a Bus                             The results will not include roads where
                                                                                      buses are prohibited. Using this restriction will also ensure that
                                                                                      the results will honor one-way streets.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Driving a Taxi                            The results will not include roads where
                                                                                      taxis are prohibited. Using this restriction will also ensure that
                                                                                      the results will honor one-way streets.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Driving a Truck                           The results will not include roads where
                                                                                      trucks are prohibited. Using this restriction will also ensure that
                                                                                      the results will honor one-way streets.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Driving an Automobile                     The results will not include roads
                                                                                      where automobiles are prohibited. Using this restriction will also
                                                                                      ensure that the results will honor one-way streets.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Driving an Emergency Vehicle              The results will not include
                                                                                      roads where emergency vehicles are prohibited. Using this
                                                                                      restriction will also ensure that the results will honor one-way
                                                                                      streets.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Height Restriction                        The results will not include roads
                                                                                      where the vehicle height exceeds the maximum allowed height for the
                                                                                      road. The vehicle height can be specified using the Vehicle Height
                                                                                      (meters) restriction parameter.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Kingpin to Rear Axle                      The results will
                                            Length Restriction                        not include roads where the vehicle length exceeds the maximum
                                                                                      allowed kingpin to rear axle for all trucks on the road. The length
                                                                                      between the vehicle kingpin and the rear axle can be specified
                                                                                      using the Vehicle Kingpin to Rear Axle Length (meters) restriction
                                                                                      parameter.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Length Restriction                        The results will not include roads
                                                                                      where the vehicle length exceeds the maximum allowed length for the
                                                                                      road. The vehicle length can be specified using the Vehicle Length
                                                                                      (meters) restriction parameter.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Preferred for Pedestrians                 The result prefers paths designated for pedestrians.
                                            ----------------------------------------  ------------------------------------------------
                                            Riding a Motorcycle                       The results will not include roads
                                                                                      where motorcycles are prohibited. Using this restriction will also
                                                                                      ensure that the results will honor one-way streets.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Roads Under Construction Prohibited       The results will not
                                                                                      include roads that are under construction.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Semi or Tractor with One                  The results will not include roads where semis or tractors with
                                            or More Trailers Prohibited               one or more trailers are prohibited.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Single Axle Vehicles Prohibited           The results will not
                                                                                      include roads where vehicles with single axles are
                                                                                      prohibited.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Tandem Axle Vehicles Prohibited           The results will not
                                                                                      include roads where vehicles with tandem axles are
                                                                                      prohibited.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Through Traffic Prohibited                The results will not include
                                                                                      roads where through traffic (non local) is prohibited.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Truck with Trailers Restriction           The results will not
                                                                                      include roads where trucks with the specified number of trailers on
                                                                                      the truck are prohibited. The number of trailers on the truck can
                                                                                      be specified using the Number of Trailers on Truck restriction
                                                                                      parameter.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Use Preferred Hazmat Routes               The results will prefer roads
                                                                                      that are designated for transporting any kind of hazardous
                                                                                      materials.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Use Preferred Truck Routes                The results will prefer roads
                                                                                      that are designated as truck routes, such as the roads that are
                                                                                      part of the national network as specified by the National Surface
                                                                                      Transportation Assistance Act in the United States, or roads that
                                                                                      are designated as truck routes by the state or province, or roads
                                                                                      that are preferred by the trucks when driving in an
                                                                                      area.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Walking                                   The results will not include roads where
                                                                                      pedestrians are prohibited.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Weight Restriction                        The results will not include roads
                                                                                      where the vehicle weight exceeds the maximum allowed weight for the
                                                                                      road. The vehicle weight can be specified using the Vehicle Weight
                                                                                      (kilograms) restriction parameter.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Weight per Axle Restriction               The results will not include
                                                                                      roads where the vehicle weight per axle exceeds the maximum allowed
                                                                                      weight per axle for the road. The vehicle weight per axle can be
                                                                                      specified using the Vehicle Weight per Axle (kilograms) restriction
                                                                                      parameter.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Width Restriction                         The results will not include roads where
                                                                                      the vehicle width exceeds the maximum allowed width for the road.
                                                                                      The vehicle width can be specified using the Vehicle Width (meters)
                                                                                      restriction parameter.
                                                                                      Availability: Select countries in North America and Europe
                                            ========================================  ================================================

                                            Choice list:['Any Hazmat Prohibited', 'Avoid Carpool Roads', 'Avoid Express Lanes', 'Avoid Ferries', 'Avoid Gates',
                                            'Avoid Limited Access Roads', 'Avoid Private Roads', 'Avoid Roads Unsuitable for Pedestrians', 'Avoid Stairways',
                                            'Avoid Toll Roads', 'Avoid Toll Roads for Trucks', 'Avoid Truck Restricted Roads', 'Avoid Unpaved Roads',
                                            'Axle Count Restriction', 'Driving a Bus', 'Driving a Delivery Vehicle', 'Driving a Taxi', 'Driving a Truck',
                                            'Driving an Automobile', 'Driving an Emergency Vehicle', 'Height Restriction', 'Kingpin to Rear Axle Length Restriction',
                                            'Length Restriction', 'Preferred for Pedestrians', 'Riding a Motorcycle', 'Roads Under Construction Prohibited',
                                            'Semi or Tractor with One or More Trailers Prohibited', 'Single Axle Vehicles Prohibited', 'Tandem Axle Vehicles Prohibited',
                                            'Through Traffic Prohibited', 'Truck with Trailers Restriction', 'Use Preferred Hazmat Routes', 'Use Preferred Truck Routes',
                                            'Walking', 'Weight Restriction', 'Weight per Axle Restriction', 'Width Restriction']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    attribute_parameter_values              Optional :class:`~arcgis.features.FeatureSet`  . Specify additional values required by some restrictions, such as the weight of a vehicle
                                            for Weight Restriction. You can also use the attribute parameter to specify whether any restriction prohibits,
                                            avoids, or prefers travel on roads that use the restriction. If the restriction is
                                            meant to avoid or prefer roads, you can further specify the degree
                                            to which they are avoided or preferred using this
                                            parameter. For example, you can choose to never use toll roads, avoid them as much as possible, or even highly
                                            prefer them.
                                            The value you provide for this parameter is ignored unless Travel Mode is set to Custom, which is the default value.
                                            If you specify the Attribute Parameter Values parameter from a feature class, the field names on the feature class must match the fields as described below:

                                            * ``AttributeName``: Lists the name of the restriction.
                                            * ``ParameterName``: Lists the name of the parameter associated with the restriction. A restriction can have one or more ParameterName field values based on its intended use.
                                            * ``ParameterValue``: The value for ``ParameterName`` used by the tool when evaluating the restriction.
                                              Attribute Parameter Values is dependent on the Restrictions parameter. The ParameterValue field is applicable only
                                              if the restriction name is specified as the value for the
                                              Restrictions parameter.
                                              In Attribute Parameter Values, each restriction (listed as AttributeName) has a ParameterName field
                                              value, Restriction Usage, that specifies whether the restriction
                                              prohibits, avoids, or prefers travel on the roads associated with
                                              the restriction and the degree to which the roads are avoided or
                                              preferred. The Restriction Usage ParameterName can be assigned any of
                                              the following string values or their equivalent numeric values
                                              listed within the parentheses:

                                              * ``PROHIBITED`` (-1) - Travel on the roads using the restriction is completely prohibited.
                                              * ``AVOID_HIGH`` (5) - It is highly unlikely for the tool to include in the route the roads that are associated with the restriction.
                                              * ``AVOID_MEDIUM`` (2) - It is unlikely for the tool to include in the route the roads that are associated with the restriction.
                                              * ``AVOID_LOW`` (1.3) - It is somewhat unlikely for the tool to include in the route the roads that are associated with the restriction.
                                              * ``PREFER_LOW`` (0.8) - It is somewhat likely for the tool to include in the route the roads that are associated with the restriction.
                                              * ``PREFER_MEDIUM`` (0.5) - It is likely for the tool to include in the route the roads that are associated with the restriction.
                                              * ``PREFER_HIGH`` (0.2) - It is highly likely for the tool to include in the route the roads.

                                              In most cases, you can use the default value, PROHIBITED,
                                              for the Restriction Usage if the restriction is dependent on a
                                              vehicle-characteristic such as vehicle height. However, in some
                                              cases, the value for Restriction Usage depends on your routing
                                              preferences. For example, the Avoid Toll Roads restriction has the
                                              default value of AVOID_MEDIUM for the Restriction Usage parameter.
                                              This means that when the restriction is used, the tool will try to
                                              route around toll roads when it can. AVOID_MEDIUM also indicates
                                              how important it is to avoid toll roads when finding the best
                                              route; it has a medium priority. Choosing AVOID_LOW would put lower
                                              importance on avoiding tolls; choosing AVOID_HIGH instead would
                                              give it a higher importance and thus make it more acceptable for
                                              the service to generate longer routes to avoid tolls. Choosing
                                              PROHIBITED would entirely disallow travel on toll roads, making it
                                              impossible for a route to travel on any portion of a toll road.
                                              Keep in mind that avoiding or prohibiting toll roads, and thus
                                              avoiding toll payments, is the objective for some; in contrast,
                                              others prefer to drive on toll roads because avoiding traffic is
                                              more valuable to them than the money spent on tolls. In the latter
                                              case, you would choose PREFER_LOW, PREFER_MEDIUM, or PREFER_HIGH as
                                              the value for Restriction Usage. The higher the preference, the
                                              farther the tool will go out of its way to travel on the roads
                                              associated with the restriction.

                                              ========================================  =========================  =======================
                                              **AttributeName**                         **ParameterName**          **ParameterValue**
                                              ----------------------------------------  -------------------------  -----------------------
                                              Any Hazmat Prohibited                     Restriction Usage           PROHIBITED

                                              ----------------------------------------  -------------------------  -----------------------
                                              Avoid Carpool Roads                       Restriction Usage          PROHIBITED
                                              ----------------------------------------  -------------------------  -----------------------
                                              Avoid Express Lanes                       Restriction Usage          PROHIBITED
                                              ----------------------------------------  -------------------------  -----------------------
                                              Avoid Ferries                             Restriction Usage          AVOID_MEDIUM
                                              ----------------------------------------  -------------------------  -----------------------
                                              Avoid Gates                               Restriction Usage          AVOID_MEDIUM
                                              ----------------------------------------  -------------------------  -----------------------
                                              Avoid Limited Access Roads                Restriction Usage          AVOID_MEDIUM
                                              ----------------------------------------  -------------------------  -----------------------
                                              Avoid Private Roads                       Restriction Usage          AVOID_MEDIUM
                                              ----------------------------------------  -------------------------  -----------------------
                                              Avoid Roads Unsuitable for Pedestrians    Restriction Usage          AVOID_HIGH
                                              Avoid Stairways                           Restriction Usage          AVOID_HIGH
                                              ----------------------------------------  -------------------------  -----------------------
                                              Avoid Toll Roads                          Restriction Usage          AVOID_MEDIUM
                                              ----------------------------------------  -------------------------  -----------------------
                                              Avoid Toll Roads for Trucks               Restriction Usage          AVOID_MEDIUM
                                              ----------------------------------------  -------------------------  -----------------------
                                              Avoid Truck Restricted Roads              Restriction Usage          AVOID_HIGH
                                              ----------------------------------------  -------------------------  -----------------------
                                              Avoid Unpaved Roads                       Restriction Usage          AVOID_HIGH
                                              ----------------------------------------  -------------------------  -----------------------
                                              Axle Count Restriction                    Number of Axles            0

                                                                                        Restriction Usage          PROHIBITED
                                              ----------------------------------------  -------------------------  -----------------------
                                              Driving a Bus                             Restriction Usage          PROHIBITED
                                              ----------------------------------------  -------------------------  -----------------------
                                              Driving a Taxi                            Restriction Usage          PROHIBITED
                                              ---------------------------------------  -------------------------  -----------------------
                                              Driving a Truck                           Restriction Usage          PROHIBITED
                                              ----------------------------------------  -------------------------  -----------------------
                                              Driving an Automobile                     Restriction Usage          PROHIBITED
                                              ----------------------------------------  -------------------------  -----------------------
                                              Driving an Emergency Vehicle              Restriction Usage          PROHIBITED
                                              ----------------------------------------  -------------------------  -----------------------
                                              Height Restriction                        Restriction Usage          PROHIBITED

                                                                                        Vehicle Height (meters)    0
                                              ----------------------------------------  -------------------------  -----------------------
                                              Kingpin to Rear Axle                      Restriction Usage          PROHIBITED
                                              Length Restriction
                                                                                        Vehicle Kingpin to Rear    0
                                                                                        Axle Length (meters)
                                              ----------------------------------------  -------------------------  -----------------------
                                              Length Restriction                        Restriction Usage          PROHIBITED
                                                                                        Vehicle Length (meters)    0
                                              ----------------------------------------  -------------------------  -----------------------
                                              Preferred for Pedestrians                 Restriction Usage          PREFER_LOW
                                              ----------------------------------------  -------------------------  -----------------------
                                              Riding a Motorcycle                       Restriction Usage          PROHIBITED
                                              ----------------------------------------  -------------------------  -----------------------
                                              Roads Under Construction Prohibited       Restriction Usage          PROHIBITED
                                              ----------------------------------------  -------------------------  -----------------------
                                              Semi or Tractor with One                  Restriction Usage          PROHIBITED
                                              or more trailers prohibited
                                              ----------------------------------------  -------------------------  -----------------------
                                              Single Axle Vehicles Prohibited           Restriction Usage          PROHIBITED
                                              ----------------------------------------  -------------------------  -----------------------
                                              Tandem Axle Vehicles Prohibited           Restriction Usage          PROHIBITED
                                              ----------------------------------------  -------------------------  -----------------------
                                              Through Traffic Prohibited                Restriction Usage          AVOID_HIGH
                                              ----------------------------------------  -------------------------  -----------------------
                                              Truck with Trailers Restriction           Restriction Usage          PROHIBITED

                                                                                        Number of Trailers         0
                                                                                        on Truck
                                              ----------------------------------------  -------------------------  -----------------------
                                              Use Preferred Hazmat Routes               Restriction Usage          PREFER_MEDIUM
                                              ----------------------------------------  -------------------------  -----------------------
                                              Use Preferred Truck Routes                Restriction Usage          PREFER_HIGH
                                              ----------------------------------------  -------------------------  -----------------------
                                              Walking                                   Restriction Usage          PROHIBITED
                                              ----------------------------------------  -------------------------  -----------------------
                                              WalkTime                                  Walking Speed (km/h)       5
                                              ----------------------------------------  -------------------------  -----------------------
                                              Weight Restriction                        Restriction Usage          PROHIBITED

                                                                                        Vehicle Weight             0
                                                                                        (kilograms)
                                              ----------------------------------------  -------------------------  -----------------------
                                              Weight per Axle Restriction               Restriction Usage          PROHIBITED

                                                                                        Vehicle Weight per          0
                                                                                        Axle (kilograms)
                                              ----------------------------------------  -------------------------  -----------------------
                                              Width Restriction                         Restriction Usage          PROHIBITED

                                                                                        Vehicle Width              0
                                                                                        (meters)
                                              ========================================  =========================  =======================
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    impedance                               Optional string.  Specify the impedance, which is a value that represents the effort or cost of traveling along
                                            road segments or on other parts of the transportation network.
                                            Travel distance is an impedance; the length of a road in kilometers can be thought of as impedance. Travel distance
                                            in this sense is the same for all modes-a kilometer for a pedestrian is also a kilometer for a car. (What may change
                                            is the pathways on which the different modes are allowed to travel, which affects distance between points, and this
                                            is modeled by travel mode settings.) Travel time can also be an impedance; a car may take one minute to travel a
                                            mile along an empty road. Travel times can vary by travel mode-a pedestrian may take more than 20  minutes to walk
                                            the same mile, so it is important to choose the right impedance for the travel mode you are modeling.
                                            Choose from the following impedance values: Drive Time-Models travel times for a car. These travel times are
                                            dynamic and fluctuate according to traffic flows in areas where traffic data is available. Truck Time-Models
                                            travel times for a truck.  These travel times are static for each road and don't fluctuate with traffic. This is
                                            the default value. Walk Time-Models travel times for a pedestrian. Travel Distance-Stores  length measurements
                                            along roads and paths. To model walk distance, choose this option and ensure Walking is  set in the Restriction
                                            parameter. Similarly, to model drive or truck distance, choose Travel Distance here and set the appropriate
                                            restrictions so your vehicle travels only on roads where it is permitted to do so. The value you provide for this
                                            parameter is ignored unless Travel Mode is set to Custom. If you choose Drive Time, Truck Time, or Walk Time, the
                                            Measurement Units parameter must be set to a time-based value; if you choose Travel Distance for Impedance,
                                            Measurement Units must be distance-based.

                                            Choice list:['Drive Time', 'Truck Time', 'Walk Time', 'Travel Distance']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    origin_destination_line_shape           Optional string. The resulting lines of an OD cost matrix can be represented with either straight-line geometry
                                            or no geometry at all. In both cases, the route is always computed along the street network by minimizing the
                                            travel time or the travel distance, never using the straight-line
                                            distance between origins and destinations.

                                            ``Straight Line``: Straight lines connect origins and destinations.
                                            ``None``: Do not return any shapes for the lines that connect origins and destinations. This is useful when you have a large number of origins and destinations and are interested only in the OD cost matrix table (and not the output line shapes).

                                            Choice list:['None', 'Straight Line']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    save_output_layer                       Optional boolean. Specify if the tool should save the analysis settings as a network analysis layer file.
                                            You cannot directly work with this file even when you open the file in an ArcGIS Desktop application like ArcMap.
                                            It is meant to be sent to Esri Technical Support to diagnose the quality of results returned from the tool.
                                            True: Save the network analysis layer file. The file is downloaded in a temporary directory on your machine. In ArcGIS Pro, the location of the downloaded file can be determined by viewing the value for the Output Network Analysis Layer parameter in the entry corresponding to the tool execution in the Geoprocessing history of your Project. In ArcMap, the location of the file can be determined by accessing the Copy Location option in the shortcut menu on the Output Network Analysis Layer parameter in the entry corresponding to the tool execution in the Geoprocessing Results window.
                                            False: Do not save the network analysis layer file. This is the default.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    overrides                               Optional string. Specify additional settings that can influence the behavior of the solver when finding solutions
                                            for the network analysis problems. The value for this parameter needs to be specified in JavaScript Object Notation
                                            (JSON). For example, a valid value is of the following form {"overrideSetting1" : "value1", "overrideSetting2" :
                                            "value2"}. The override setting name is always enclosed in double quotes. The values can be a number, Boolean,
                                            or string. The default value for this parameter is no value, which indicates not to override any solver settings. Overrides
                                            are advanced settings that should be used only after careful analysis of the results obtained before and after applying
                                            the settings. A list of supported override settings for each solver and their acceptable values can be obtained by contacting
                                            Esri Technical Support.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    save_route_data                         Optional boolean. Choose whether the output includes a zip file that contains a file geodatabase holding the inputs
                                            and outputs of the analysis in a format that can be used to share route layers with ArcGIS Online or Portal for
                                            ArcGIS.
                                            True: Save the route data as a zip file. The file is downloaded in a temporary directory on your machine. In ArcGIS Pro, the location of the downloaded file can be determined by viewing the value for the Output Route Data parameter in the entry corresponding to the tool execution in the Geoprocessing history of your Project. In ArcMap, the location of the file can be determined by accessing the Copy Location option in the shortcut menu on the Output Route Data parameter in the entry corresponding to the tool execution in the Geoprocessing Results window.
                                            False: Do not save the route data. This is the default.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    time_impedance                          Optional string. Specify the time-based impedance.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    distance_impedence                      Optional string. Specify the distance-based impedance.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    output_format                           Optional. Specify the format in which the output features are created.

                                            Choose from the following formats:

                                            * Feature Set - The output features are returned as feature classes and tables. This is the default.
                                            * JSON File - The output features are returned as a compressed file containing the JSON representation of the outputs. When this option is specified, the output is a single file (with a .zip extension) that contains one or more JSON files (with a .json extension) for each of the outputs created by the service.
                                            * GeoJSON File - The output features are returned as a compressed file containing the GeoJSON representation of the outputs. When this option is specified, the output is a single file (with a .zip extension) that contains one or more GeoJSON files (with a .geojson extension) for each of the outputs created by the service.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    gis                                     Optional, the :class:`~arcgis.gis.GIS` on which this tool runs. If not specified, the active GIS is used.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    future                                  Optional boolean. If True, a future object will be returned and the process
                                            will not wait for the task to complete. The default is False, which means wait for results.
    ======================================  ==========================================================================================================================================

    :returns: the following as a named tuple:

    * solve_succeeded - Solve Succeeded as a bool
    * output_origin_destination_lines - Output Origin Destination Lines as a FeatureSet
    * output_origins - Output Origins as a FeatureSet
    * output_destinations - Output Destinations as a FeatureSet

    Click `GenerateOriginDestinationCostMatrix <https://developers.arcgis.com/rest/network/api-reference/origin-destination-cost-matrix-service.htm>`_ for additional help.
    """

    if gis is None:
        gis = arcgis.env.active_gis
    url = gis.properties.helperServices.asyncODCostMatrix.url
    url = _validate_url(url, gis)
    tbx = _create_toolbox(url, gis=gis)
    defaults = dict(
        zip(
            tbx.generate_origin_destination_cost_matrix.__annotations__.keys(),
            tbx.generate_origin_destination_cost_matrix.__defaults__,
        )
    )
    if origins is None:
        origins = defaults["origins"]
    if destinations is None:
        destinations = defaults["destinations"]
    if point_barriers is None:
        point_barriers = defaults["point_barriers"]
    if line_barriers is None:
        line_barriers = defaults["line_barriers"]
    if polygon_barriers is None:
        polygon_barriers = defaults["polygon_barriers"]
    if restrictions is None:
        restrictions = defaults["restrictions"]
    if attribute_parameter_values is None:
        attribute_parameter_values = defaults["attribute_parameter_values"]
    if overrides is None:
        overrides = defaults["overrides"]
    if time_impedance is None and "time_impedance" in defaults:
        time_impedance = defaults["time_impedance"]
    if distance_impedance is None and "distance_impedance" in defaults:
        distance_impedance = defaults["distance_impedance"]
    if output_format is None and "output_format" in defaults:
        output_format = defaults["output_format"]
    from arcgis._impl.common._utils import inspect_function_inputs

    params = {
        "origins": origins,
        "destinations": destinations,
        "travel_mode": travel_mode,
        "distance_units": distance_units,
        "analysis_region": analysis_region,
        "number_of_destinations_to_find": number_of_destinations_to_find,
        "cutoff": cutoff,
        "time_of_day": time_of_day,
        "time_zone_for_time_of_day": time_zone_for_time_of_day,
        "point_barriers": point_barriers,
        "line_barriers": line_barriers,
        "polygon_barriers": polygon_barriers,
        "uturn_at_junctions": uturn_at_junctions,
        "use_hierarchy": use_hierarchy,
        "restrictions": restrictions,
        "attribute_parameter_values": attribute_parameter_values,
        "impedance": impedance,
        "origin_destination_line_shape": origin_destination_line_shape,
        "save_output_network_analysis_layer": save_output_network_analysis_layer,
        "overrides": overrides,
        "time_impedance": time_impedance,
        "distance_impedance": distance_impedance,
        "output_format": output_format,
        "gis": gis,
        "future": True,
    }
    params = inspect_function_inputs(
        tbx.generate_origin_destination_cost_matrix, **params
    )
    params["future"] = True
    job = tbx.generate_origin_destination_cost_matrix(**params)
    if future:
        return job
    return job.result()


generate_origin_destination_cost_matrix.__annotations__ = {
    "origins": FeatureSet,
    "destinations": FeatureSet,
    "travel_mode": str,
    "time_units": str,
    "distance_units": str,
    "analysis_region": str,
    "number_of_destinations_to_find": int,
    "cutoff": float,
    "time_of_day": datetime,
    "time_zone_for_time_of_day": str,
    "point_barriers": FeatureSet,
    "line_barriers": FeatureSet,
    "polygon_barriers": FeatureSet,
    "uturn_at_junctions": str,
    "use_hierarchy": bool,
    "restrictions": str,
    "attribute_parameter_values": FeatureSet,
    "impedance": str,
    "origin_destination_line_shape": str,
    "return": tuple,
}



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# giving error in loss of gfl
# gfl_x101_32x4d_fpn_dconv_c4-c5_mstrain_2x_coco.py, box AP=48.1
model = dict(
    type="GFL",
    pretrained="open-mmlab://resnext101_32x4d",
    backbone=dict(
        type="ResNeXt",
        depth=101,
        groups=32,
        base_width=4,
        num_stages=4,
        out_indices=(0, 1, 2, 3),
        frozen_stages=1,
        norm_cfg=dict(type="BN", requires_grad=True),
        dcn=dict(type="DCN", deform_groups=1, fallback_on_stride=False),
        stage_with_dcn=(False, False, True, True),
        norm_eval=True,
        style="pytorch",
    ),
    neck=dict(
        type="FPN",
        in_channels=[256, 512, 1024, 2048],
        out_channels=256,
        start_level=1,
        add_extra_convs="on_output",
        num_outs=5,
    ),
    bbox_head=dict(
        type="GFLHead",
        num_classes=80,
        in_channels=256,
        stacked_convs=4,
        feat_channels=256,
        anchor_generator=dict(
            type="AnchorGenerator",
            ratios=[1.0],
            octave_base_scale=8,
            scales_per_octave=1,
            strides=[8, 16, 32, 64, 128],
        ),
        loss_cls=dict(
            type="QualityFocalLoss", use_sigmoid=True, beta=2.0, loss_weight=1.0
        ),
        loss_dfl=dict(type="DistributionFocalLoss", loss_weight=0.25),
        reg_max=16,
        loss_bbox=dict(type="GIoULoss", loss_weight=2.0),
    ),
    # training and testing settings
    train_cfg=dict(
        assigner=dict(type="ATSSAssigner", topk=9),
        allowed_border=-1,
        pos_weight=-1,
        debug=False,
    ),
    test_cfg=dict(
        nms_pre=1000,
        min_bbox_size=0,
        score_thr=0.05,
        nms=dict(type="nms", iou_threshold=0.6),
        max_per_img=100,
    ),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/gfl/gfl_x101_32x4d_fpn_dconv_c4-c5_mstrain_2x_coco/gfl_x101_32x4d_fpn_dconv_c4-c5_mstrain_2x_coco_20200630_102002-14a2bf25.pth"



# MIT License

# Copyright (c) 2019 Hengshuang Zhao

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Based on https://github.com/hszhao/semseg

import torch
import numpy as np
import torch.nn.functional as F
import torch.nn as nn
import torch
import math
from fastai.callbacks.hooks import hook_outputs
from fastai.callbacks.hooks import model_sizes
from torchvision.models.segmentation.deeplabv3 import DeepLabHead
from torchvision.models.segmentation.fcn import FCNHead
from ._timm_utils import get_backbone
from fastprogress.fastprogress import progress_bar
from ._PointRend import PointRendSemSegHead
from fastai.vision import flatten_model


def get_dilation_index(backbone_name, pointrend=False, keep_dilation=False):
    vgg = False
    if "vgg" in backbone_name:
        modify_dilation_index = -5
        vgg = True
    elif "mobilenet" in backbone_name or "hardcorenas" in backbone_name:
        modify_dilation_index = -1
    else:
        if pointrend and not keep_dilation:
            modify_dilation_index = -1
        else:
            modify_dilation_index = -2

    return modify_dilation_index, vgg


def get_last_module(backbone):
    hookable_modules = list(backbone.children())
    if len(hookable_modules) < 5:

        def get_len(m):
            try:
                return len(m)
            except:
                return 0

        module_len = [get_len(m) for m in hookable_modules]
        noise = np.array(range(len(module_len))) * 1e-15
        module_len += noise
        hookable_modules = hookable_modules[np.argmax(module_len)]

    return hookable_modules


def get_hooks(backbone_name, hookable_modules):
    if "vgg" in backbone_name:
        hooks = [
            hookable_modules[i - 1]
            for i, module in enumerate(hookable_modules)
            if isinstance(module, nn.MaxPool2d)
        ]

    else:
        hooks = [hookable_modules[-2], hookable_modules[-4]]

    return hooks


def add_dilation(backbone_fn, hookable_modules, modify_dilation_index):
    custom_idx = 0
    for i, module in enumerate(hookable_modules[modify_dilation_index:]):
        dilation = 2 * (i + 1)
        padding = 2 * (i + 1)

        if "vgg" in backbone_fn.__name__:
            if isinstance(module, nn.Conv2d):
                dilation = 2 * (custom_idx + 1)
                padding = 2 * (custom_idx + 1)
                module.dilation, module.padding, module.stride = (
                    (dilation, dilation),
                    (padding, padding),
                    (1, 1),
                )
                custom_idx += 1
        else:
            for m in flatten_model(module):
                if "Pool" in m.__class__.__name__:
                    m.stride = 1
                    m.kernel_size = 3
                    m.padding = 1

                elif hasattr(m, "stride") and len(m.stride) == 2:
                    if m.kernel_size[0] > 1:
                        m.dilation, m.padding, m.stride = (
                            (dilation, dilation),
                            (padding, padding),
                            (1, 1),
                        )
                    else:
                        m.stride = (1, 1)


class Deeplab(nn.Module):
    def __init__(
        self,
        num_classes,
        backbone_fn,
        chip_size=224,
        pointrend=True,
        keep_dilation=False,
        pretrained=True,
    ):
        super().__init__()
        self.pointrend = pointrend
        self.backbone = get_backbone(backbone_fn, pretrained)
        backbone_name = backbone_fn.__name__
        modify_dilation_index, self.vgg = get_dilation_index(
            backbone_name, pointrend, keep_dilation
        )
        hookable_modules = get_last_module(self.backbone)
        add_dilation(backbone_fn, hookable_modules, modify_dilation_index)
        hooks = get_hooks(backbone_name, hookable_modules)

        ## Hook at the index where we need to get the auxillary logits out along with Fine-grained features
        self.hook = hook_outputs(hooks)

        ## returns the size of various activations
        feature_sizes = model_sizes(self.backbone, size=(chip_size, chip_size))

        if not self.vgg:
            ## Geting the number of channel persent in stored activation inside of the hook
            num_channels_aux_classifier = self.hook[0].stored.shape[1]
            ## Get number of channels in the last layer
            num_channels_classifier = feature_sizes[-1][1]
        else:
            num_channels_aux_classifier = self.hook[-2].stored.shape[1]
            num_channels_classifier = self.hook[-1].stored.shape[1]

        self.classifier = DeepLabHead(num_channels_classifier, num_classes)
        self.aux_classifier = FCNHead(num_channels_aux_classifier, num_classes)

        if self.pointrend:
            if self.vgg:
                num_channels = (
                    self.hook[-3].stored.shape[1] + self.hook[-4].stored.shape[1]
                )
                stride = chip_size / self.hook[-1].stored.shape[2]
            else:
                num_channels = self.hook[1].stored.shape[1]
                stride = chip_size / feature_sizes[-1][2]

            subdivision_steps = math.ceil(math.log(stride, 2))
            self.pointrend_head = PointRendSemSegHead(
                num_classes,
                num_channels,
                train_num_points=(chip_size / stride) ** 2,
                subdivision_num_points=(chip_size / (stride / 2)) ** 2,
                subdivision_steps=subdivision_steps,
            )

    def forward(self, x):
        x_size = x.size()
        x = self.backbone(x)
        features = self.hook.stored

        if self.vgg:
            x = self.classifier(features[-1])
        else:
            x = self.classifier(x)

        if self.pointrend:
            if self.vgg:
                pointrend_out = self.pointrend_head(x, [features[-4], features[-3]])
            else:
                pointrend_out = self.pointrend_head(x, [features[1]])

        result = F.interpolate(x, x_size[2:], mode="bilinear", align_corners=False)

        if self.training:
            if self.vgg:
                x = self.aux_classifier(features[-2])
            else:
                x = self.aux_classifier(features[0])

            x = F.interpolate(x, x_size[2:], mode="bilinear", align_corners=False)

            if self.pointrend:
                return result, x, pointrend_out
            else:
                return result, x
        else:
            if self.pointrend:
                if pointrend_out.shape[-1] != x_size[-1]:
                    pointrend_out = F.interpolate(
                        pointrend_out, x_size[2:], mode="bilinear", align_corners=True
                    )
                return pointrend_out
            else:
                return result


def mask_iou(mask1, mask2):
    mask1 = mask1.permute(0, 2, 3, 1)
    mask2 = mask2.permute(0, 2, 3, 1)
    mask1 = torch.reshape(mask1 > 0, (-1, mask1.shape[-1])).type(torch.float64)
    mask2 = torch.reshape(mask2 > 0, (-1, mask2.shape[-1])).type(torch.float64)
    area1 = torch.sum(mask1, dim=0)
    area2 = torch.sum(mask2, dim=0)
    intersection = torch.sum(mask1 * mask2, dim=0)
    union = area1 + area2 - intersection
    iou = intersection / (union + 1e-6)
    return iou


def compute_miou(model, dl, mean, num_classes, show_progress, ignore_mapped_class=[]):
    ious = []
    model.learn.model.eval()
    with torch.no_grad():
        for input, target in progress_bar(dl, display=show_progress):
            if getattr(model, "_is_model_extension", False):
                if model._is_multispectral:
                    pred = model.learn.model(
                        model._model_conf.transform_input_multispectral(input)
                    )
                else:
                    pred = model.learn.model(model._model_conf.transform_input(input))
            else:
                pred = model.learn.model(input)
            target = target.squeeze(1)
            if ignore_mapped_class != []:
                for k in ignore_mapped_class:
                    pred[:, k] = pred.min() - 1
                pred = pred.argmax(dim=1)
            else:
                pred = pred.argmax(dim=1)
            mask1 = []
            mask2 = []
            for i in range(pred.shape[0]):
                mask1.append(
                    pred[i].to(model._device)
                    == num_classes[:, None, None].to(model._device)
                )
                mask2.append(
                    target[i].to(model._device)
                    == num_classes[:, None, None].to(model._device)
                )
            mask1 = torch.stack(mask1)
            mask2 = torch.stack(mask2)
            iou = mask_iou(mask1, mask2)
            ious.append(iou.tolist())

    return np.mean(ious, 0)



import os
from arcgis._impl.common._isd import InsensitiveDict
from typing import List, Dict, Any
from arcgis.gis import GIS


###########################################################################
class NotebookFile:
    """Represents a Single File on the ArcGIS Notebook Server"""

    _da = None
    _definition = None

    # ---------------------------------------------------------------------
    def __init__(self, definition: Dict[str, Any], da: "NotebookDataAccess"):
        self._definition = definition
        self._da = da

    # ---------------------------------------------------------------------
    def __str__(self):
        return f"<NotebookFile file={self.properties.name}>"

    # ---------------------------------------------------------------------
    def __repr__(self):
        return f"<NotebookFile file={self.properties.name}>"

    # ---------------------------------------------------------------------
    @property
    def properties(self):
        """
        returns the properties of the resource

        :return: Dict

        """
        return InsensitiveDict(self._definition)

    # ---------------------------------------------------------------------
    def download(self) -> str:
        """
        Copies down the data from the server to the local machine

        :return: str as file path
        """
        return self._da._download(filename=self.properties["Name"])

    # ---------------------------------------------------------------------
    def erase(self) -> bool:
        """
        Deletes a file from the system

        :return: Boolean
        """
        return self._da._delete(filename=self.properties["Name"])


###########################################################################
class NotebookDataAccess:
    """
    The Data Access Workspace Directory allows notebook authors to manage files used in their notebooks.
    """

    _url = None
    _gis = None

    # ---------------------------------------------------------------------
    def __init__(self, url, gis):
        self._url = url
        self._gis = gis

    # ---------------------------------------------------------------------
    def upload(self, fp: str) -> bool:
        """
        Uploads a file to the Notebook Server

        ===================  ==========================================================================
        **Parameter**         **Description**
        -------------------  --------------------------------------------------------------------------
        fp                   Required String. The path of the file to upload
        ===================  ==========================================================================

        :return: Boolean
        """

        url = f"{self._url}/notebookworkspace/{os.path.basename(fp)}"
        if os.path.isfile(fp) == False:
            raise ValueError(f"Cannot find file: {fp}")
        additional_headers = {
            "Content-Type": "application/octet-stream",
            "Content-Length": f"{os.path.getsize(fp)}",
            "x-ms-blob-type": "BlockBlob",
            "x-ms-version": "2020-02-10",
        }
        resp = self._gis._con.put_raw(
            url, data=open(fp, "rb"), additional_headers=additional_headers
        )
        return resp.status_code >= 200 and resp.status_code < 300

    # ---------------------------------------------------------------------
    @property
    def files(self) -> List[Dict[str, Any]]:
        """
        Lists files that are located in the workspace directory (/arcgis/home) of the user making the request.

        :return: List[Dict[str, Any]]
        """
        url = f"{self._url}/notebookworkspace"
        params = {
            "f": "json",
            "restype": "container",
            "comp": "list",
            "token": self._gis._con.token,
        }
        return [
            NotebookFile(f, self)
            for f in self._gis._con.get(url, params).pop("Blobs", [])
        ]

    # ---------------------------------------------------------------------
    def _download(self, filename: str) -> str:
        """
        downloads a file from the
        """
        url = f"{self._url}/notebookworkspace/downloadFile"
        params = {
            "f": "json",
            "fileName": filename,
        }
        return self._gis._con.post(url, params)

    # ---------------------------------------------------------------------
    def _delete(self, filename: str) -> bool:
        """
        downloads a file from the
        """
        url = f"{self._url}/notebookworkspace/deleteFile"
        params = {
            "f": "json",
            "fileName": filename,
        }
        return self._gis._con.post(url, params).get("status") == "success"



from sklearn.metrics import (
    f1_score,
    accuracy_score,
    precision_score,
    recall_score,
    matthews_corrcoef,
    roc_auc_score,
    confusion_matrix,
    classification_report,
    r2_score,
    mean_squared_error,
    mean_absolute_error,
    mean_absolute_percentage_error,
)

from fairlearn.metrics import (
    MetricFrame,
    selection_rate,
    equalized_odds_difference,
    equalized_odds_ratio,
    demographic_parity_difference,
    demographic_parity_ratio,
    false_positive_rate,
    false_negative_rate,
    count,
    accuracy_score_ratio,
)
from fairlearn.reductions import (
    BoundedGroupLoss,
    ZeroOneLoss,
    DemographicParity,
    ErrorRate,
    GridSearch,
)

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import matplotlib.cm as cm
from matplotlib.colors import Normalize
import matplotlib.patches as mpatches
import warnings


def score(
    _is_classification,
    y_true,
    y_pred,
):
    if _is_classification:
        return accuracy_score(y_true, y_pred)
    else:
        return mean_absolute_error(y_true, y_pred)


def calculate_metrics(
    is_classification,
    data,
    y_true,
    y_pred,
    group_test,
    sensitive_feature,
    fairness_metrics,
    visualize,
):
    if not is_classification:
        fairness_ratio_threshold = 0.7
        fairness_diff_threshold = 0.01
        if fairness_metrics is None:
            fairness_metrics = "RMSE"

        return get_regression_metrics(
            data,
            y_true,
            y_pred,
            group_test,
            fairness_metrics,
            visualize,
            fairness_ratio_threshold,
            fairness_diff_threshold,
        )
    else:
        return show_classification_score(
            data,
            y_true,
            y_pred,
            group_test,
            sensitive_feature,
            fairness_metrics,
            visualize,
        )


def get_regression_metrics(
    _data,
    y_test,
    y_pred,
    df_sensitive_feature,
    metric_name,
    visualize=True,
    fairness_ratio_threshold=0.7,
    fairness_diff_threshold=0.01,
):
    for col in df_sensitive_feature.columns:
        mdf = get_mdf(
            _data,
            df_sensitive_feature,
            col,
            y_test,
            y_pred,
        )

        metric_name = metric_name.upper()

        privileged_value = mdf.index[mdf[metric_name][1:].argmax() + 1]
        underprivileged_value = mdf.index[mdf[metric_name][1:].argmin() + 1]

        metric_min = mdf[metric_name].loc[privileged_value]
        metric_max = mdf[metric_name].loc[underprivileged_value]

        ratio = np.round(metric_min / metric_max, 4)
        diff = np.round(metric_max - metric_min, 4)

        is_ratio_fair = False
        is_diff_fair = False

        fairness_metric_ratio = ratio
        if ratio > fairness_ratio_threshold:
            is_ratio_fair = True

        fairness_metric_diff = diff
        if diff < fairness_diff_threshold:
            is_diff_fair = True

        try:
            if visualize:
                fig, ax = plt.subplots()

                xlabels = mdf.index.to_list()
                yy = mdf.loc[:, metric_name]
                ax.bar(xlabels, yy, align="center", tick_label=xlabels)

                ax.set_facecolor("blanchedalmond")
                ax.set_alpha(0.7)
                plt.xticks(rotation=45)

                ax.set_xlabel(col)
                ax.set_ylabel(metric_name)

                plt.show()

        except Exception as ex:
            print(f"Exception occured : {ex}")

        fairness_metrics = {}

        fairness_metrics[col] = {
            "fairness_metric_name": metric_name,
            "fairness_metric_diff": fairness_metric_diff,
            "fairness_metric_ratio": fairness_metric_ratio,
            "is_ratio_fair": is_ratio_fair,
            "is_diff_fair": is_diff_fair,
        }

        return pd.DataFrame(fairness_metrics).T


def get_mdf(_data, sensitive_features, col, y_test, y_pred):
    regression_metrics = {
        "MAE": mean_absolute_error,
        "MSE": mean_squared_error,
        "RMSE": lambda t, p, sample_weight=None: np.sqrt(
            mean_squared_error(t, p, sample_weight=sample_weight)
        ),
        "R2": r2_score,
        "MAPE": mean_absolute_percentage_error,
    }

    overall = {}

    for k, v in regression_metrics.items():
        overall[k] = v(y_test, y_pred)

    values = sensitive_features[col].unique()
    _encoder = _data._encoder_mapping[col]
    ix = np.asarray(values.astype(int).tolist()).reshape(-1, 1)
    labels = np.unique(_encoder.inverse_transform(ix)).tolist()

    all_metrics = [overall]

    for value in values:
        metrics = {}
        for k, v in regression_metrics.items():
            metrics[k] = v(
                y_test[sensitive_features[col] == value],
                y_pred[sensitive_features[col] == value],
            )
        all_metrics += [metrics]

    mdf = pd.DataFrame(all_metrics, index=["Overall"] + labels)
    return mdf


def show_classification_score(
    data, y_true, y_pred, group_test, sensitive_feature, fairness_metrics, visualize
):
    metrics = {
        "accuracy": accuracy_score,
        "false positive rate": false_positive_rate,
        "false negative rate": false_negative_rate,
        "selection rate": selection_rate,
        "count": count,
    }

    fairness_dict = {
        "equalized_odds_difference": equalized_odds_difference,
        "demographic_parity_difference": demographic_parity_difference,
        "equalized_odds_ratio": equalized_odds_ratio,
        "demographic_parity_ratio": demographic_parity_ratio,
    }

    if fairness_metrics is None:
        fairness_metrics = fairness_dict.keys()
    else:
        fairness_metrics = [fairness_metrics]

    mf = MetricFrame(
        metrics=metrics, y_true=y_true, y_pred=y_pred, sensitive_features=group_test
    )
    res = mf.by_group

    _encoder = data._encoder_mapping[sensitive_feature]

    ix = np.asarray(res.index.astype(int).tolist())
    res.index = _encoder.inverse_transform(ix.reshape(-1, 1))

    if visualize:
        plot_accuracy_metrics(res)

    results = (res,)

    if visualize:
        fig = plt.figure(figsize=(12, 9))

    res_summary = {}

    for num, fm in enumerate(fairness_metrics):
        is_diff = False
        try:
            _metrics = fairness_dict[fm]

            if "diff" in fm:
                is_diff = True
                thre = 0.25
            else:
                thre = 0.8

            val = _metrics(y_true, y_pred, sensitive_features=group_test)
        except ZeroDivisionError as zerror:
            raise Exception(
                "One or more metrics count is zero. Please check the classification data."
            )
        val = round(val, 2)
        sum_text = get_text(fm, val, thre, is_diff)

        if visualize:
            try:
                ax = fig.add_subplot(2, 2, num + 1)
                ax.set_title(fm)

                plot_diff(fig, ax, val, thre, fm, sum_text, is_diff)
                fig.tight_layout(pad=4)
            except Exception as ex:
                warnings.warn(f"Visualization cannot be completed. {ex}")

        text_summary = " ".join(sum_text)
        res_summary[fm] = (val, text_summary)

    results = results + (res_summary,)

    return results


def show_regression_score(
    _data,
    y_true,
    y_pred,
    group_test,
    sensitive_feature,
    fairness_metrics,
    visualize=False,
):
    if fairness_metrics is None:
        fairness_metrics = "mean_absolute_error"

    mae_frame = MetricFrame(
        metrics=mean_absolute_error,
        y_true=y_true,
        y_pred=y_pred,
        sensitive_features=group_test,
    )

    if visualize:
        _encoder = _data._encoder_mapping[sensitive_feature]

        res_idx = mae_frame.by_group.index
        ix = np.asarray(res_idx.astype(int).tolist()).reshape(-1, 1)
        labels = np.unique(_encoder.inverse_transform(ix)).tolist()

        plot_regression_metrics(
            mae_frame.by_group, labels, fairness_metrics, sensitive_feature
        )

    return mae_frame.by_group


def plot_regression_metrics(res, xlabels, metric_name, _sensitive_feature):
    fig, ax = plt.subplots()

    ax.bar(res.index, res.values, align="center", tick_label=xlabels)

    ax.set_facecolor("blanchedalmond")
    ax.set_alpha(0.7)
    plt.xticks(rotation=45)

    ax.set_xlabel(_sensitive_feature)
    ax.set_ylabel(metric_name)

    plt.show()


def plot_accuracy_metrics(res):
    _sum = res["count"].sum()

    res["count"] = res["count"].apply(lambda x: x / _sum)
    ax = res.plot(kind="bar", figsize=(6.4, 3.5))
    ax.axhspan(0, 1, facecolor="blanchedalmond", alpha=0.15)

    plt.legend(bbox_to_anchor=(1.05, 1.0), loc="upper left")
    plt.tight_layout()
    plt.show()


def get_text(fm, val, thre, is_diff):
    texts = [f"The value of {fm} is {val}"]
    if val > thre:
        texts.append(f"which is more than minimum threshold {thre}.")
    else:
        texts.append(f"which is less than minimum threshold {thre}.")

    if is_diff:
        texts.append("The ideal value of this metric is 0.")
        texts.append(f"Fairness for this metric is between 0 and {thre}.")
    else:
        texts.append("The ideal value of this metric is 1.")
        texts.append(f"Fairness for this metric is between {thre} and 1.")
    return texts


def plot_diff(fig, ax, val, thre, label, summary, is_diff):
    text_summary = "\n".join(summary)

    ax.axhspan(-1, 1, facecolor="blanchedalmond", alpha=0.15)

    red_patch = mpatches.Patch(color="darksalmon", label="Biased")
    green_patch = mpatches.Patch(color="paleturquoise", label="Fair")

    if is_diff:
        ax.set_ylim(-1, 1)

        thre = 0.25
        if val > thre:
            ax.bar(
                [1, 2],
                [0, val],
                width=0.35,
                tick_label=["", label],
                color=("darksalmon"),
            )
        else:
            ax.bar(
                [1, 2],
                [0, val],
                width=0.35,
                tick_label=["", label],
                color=("paleturquoise"),
            )

        plt.axhline(y=0, color="grey", linestyle="-")
        plt.axhline(y=thre, color="grey", linestyle="--")
        plt.axhline(y=-thre, color="grey", linestyle="--")

        plt.text(2.72, 0, "Fair", fontsize=10)
        plt.text(2.72, thre, "Bias", fontsize=10)
        plt.text(2.72, -thre, "Bias", fontsize=10)

        ax.annotate(
            "",
            xy=(1.15, 0.30),
            xycoords="axes fraction",
            xytext=(1.15, 0.70),
            arrowprops=dict(arrowstyle="<->", color="black"),
        )

        text_widget = ax.text(
            1, -1.6, text_summary, fontsize=12, color="black", ha="left", va="bottom"
        )
        text_widget.set_fontstyle("italic")

        plt.legend(handles=[red_patch, green_patch], loc="upper left")

    else:
        ax.set_ylim(0, 1)

        thre = 0.8

        if val < thre:
            ax.bar(
                [1, 2],
                [0, val],
                width=0.35,
                tick_label=["", label],
                color=("darksalmon"),
            )
        else:
            ax.bar(
                [1, 2],
                [0, val],
                width=0.35,
                tick_label=["", label],
                color=("paleturquoise"),
            )

        plt.axhline(y=0, color="grey", linestyle="-")
        plt.axhline(y=thre, color="grey", linestyle="--")

        plt.text(2.72, 0, "Bias", fontsize=10)
        plt.text(2.72, thre, "Fair", fontsize=10)

        ax.annotate(
            "",
            xy=(1.15, 0.30),
            xycoords="axes fraction",
            xytext=(1.15, 0.70),
            arrowprops=dict(arrowstyle="<-", color="black"),
        )

        text_widget = ax.text(
            1, -0.4, text_summary, fontsize=12, color="black", ha="left", va="bottom"
        )
        text_widget.set_fontstyle("italic")

        plt.legend(handles=[red_patch, green_patch], loc="lower left")

    ax.set_xlim(1, 2.7)
    ax.set_xticks([])



import os
from typing import Optional
from arcgis.gis import GIS, Item
from arcgis._impl.common._mixins import PropertyMap
import concurrent.futures


########################################################################
class NotebookManager(object):
    """
    Provides access to managing a site's notebooks. An object of this
    class can be created using :attr:`~arcgis.gis.nb.NotebookServer.notebooks` property of the
    :class:`~arcgis.gis.nb.NotebookServer` class
    """

    _url = None
    _gis = None
    _properties = None
    _nbs = None
    _snapshot = None

    # ----------------------------------------------------------------------
    def __init__(self, url, gis, nbs):
        """Constructor"""
        self._url = url
        self._nbs = nbs
        if isinstance(gis, GIS):
            self._gis = gis
            self._con = self._gis._con
        else:
            raise ValueError("Invalid GIS object")

    # ----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        try:
            params = {"f": "json"}
            res = self._gis._con.get(self._url, params)
            self._properties = PropertyMap(res)
        except:
            self._properties = PropertyMap({})

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< NotebookManager @ {url} >".format(url=self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< NotebookManager @ {url} >".format(url=self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties of the resource"""
        if self._properties is None:
            self._init()
        return self._properties

    # ----------------------------------------------------------------------
    def list(self):
        """
        Returns a list of notebook instances on the Notebook Server

        :return: List of :class:`~arcgis.gis.nb.Notebook` objects

        """
        return [
            Notebook(url=self._url, item_id=nbs["id"], properties=nbs)
            for nbs in self.properties.notebooks
        ]

    # ----------------------------------------------------------------------
    @property
    def runtimes(self):
        """
        Returns a list of all runtimes

        :return: List
        """
        url = self._url + "/runtimes"
        params = {"f": "json"}
        res = self._con.get(url, params)
        if "runtimes" in res:
            return [
                Runtime(url=url + "/{rid}".format(rid=r["id"]), gis=self._gis)
                for r in res["runtimes"]
            ]
        return []

    # ----------------------------------------------------------------------
    @property
    def snapshots(self):
        """
        Provides access to managing Notebook's snapshots

        :return: :class:`~arcgis.gis.nb.SnapshotManager`
        """
        if self._snapshot is None:
            from ._snapshot import SnapshotManager

            url = self._url + "/snapshots"
            self._snapshot = SnapshotManager(url=url, gis=self._gis)
        return self._snapshot

    # ----------------------------------------------------------------------
    def restore_runtime(self):
        """
        This operation restores the two default notebook runtimes in ArcGIS
        Notebook Server - ArcGIS Notebook Python 3 Standard and ArcGIS
        Notebook Python 3 Advanced - to their original settings.
        """
        url = self._url + "/runtimes/restore"
        params = {"f": "json"}
        res = self._con.post(url, params)
        if "status" in res:
            return res["status"] == "success"
        return res

    # ----------------------------------------------------------------------
    @staticmethod
    def _future_job(
        fn,
        task_name,
        jobid=None,
        task_url=None,
        notify=False,
        gis=None,
        **kwargs,
    ):
        """
        runs the job asynchronously

        :return: Job object
        """
        from arcgis._impl._async.jobs import Job

        tp = concurrent.futures.ThreadPoolExecutor(1)
        try:
            future = tp.submit(fn=fn, **kwargs)
        except:
            future = tp.submit(fn, **kwargs)
        tp.shutdown(False)
        return Job(future, task_name, jobid, task_url, notify, gis=gis)

    # ----------------------------------------------------------------------
    def execute_notebook(
        self,
        item: Item,
        update_portal_item: bool = True,
        parameters: Optional[list] = None,
        save_parameters: bool = False,
        future: bool = False,
    ):
        """

        The Execute Notebook operation allows administrators and users with
        the `Create and Edit Notebooks` privilege to remotely
        run a notebook that they own.  The notebook pecified in the operation will be run with all
        cells in order.

        Using this operation, you can schedule the execution of a notebook,
        either once or with a regular occurrence. This allows you to
        automate repeating tasks such as data collection and cleaning,
        content updates, and portal administration. On Linux machines, use
        a cron job to schedule the executeNotebook operation; on Windows
        machines, you can use the Task Scheduler app.

        .. note::
            To run this operation in ArcGIS Enterprise, you must log in with
            an Enterprise account. You cannot execute notebooks using the
            ArcGIS Notebook Server primary site administrator account.

        You can specify parameters to be used in the notebook at execution
        time. If you've specified one or more parameters, they'll be
        inserted into the notebook as a new cell. This cell will be placed
        at the beginning of the notebook, unless you have added the tag
        parameters to a cell.

        ====================    ====================================================================
        **Parameter**            **Description**
        --------------------    --------------------------------------------------------------------
        item                    Required :class:`~arcgis.gis.Item`. Opens an existing portal item.
        --------------------    --------------------------------------------------------------------
        update_portal_item      Optional Boolean. Specifies whether you want to update the
                                notebook's portal item after execution. The default is true. You may
                                want to specify true when the notebook you're executing contains
                                information that needs to be updated, such as a workflow that
                                collects the most recent version of a dataset. It may not be
                                important to update the portal item if the notebook won't store any
                                new information after executing, such as an administrative notebook
                                that emails reminders to inactive users.
        --------------------    --------------------------------------------------------------------
        parameters              Optional List. An optional array of parameters to add to the
                                notebook for this execution. The parameters will be inserted as a
                                new cell directly after the cell you have tagged ``parameters``.
                                Separate parameters with a comma. Use the format "x":1 when
                                defining parameters with numbers, and "y":"text" when defining
                                parameters with text strings.
        --------------------    --------------------------------------------------------------------
        save_parameters         Optional Boolean.  Specifies whether the notebook parameters cell
                                should be saved in the notebook for future use. The default is
                                false.
        --------------------    --------------------------------------------------------------------
        future                  Optional boolean. If True, a Job object will be returned and the process
                                will not wait for the task to complete. The default is False, which means wait for results.
        ====================    ====================================================================

        :return: Dict else If ``future = True``, then the result is
                 a `concurrent.futures.Future <https://docs.python.org/3/library/concurrent.futures.html>`_ object.
                 Call ``result()`` to get the response

        .. code-block:: python

            # Usage Example:

            >>> from arcgis.gis import GIS
            >>> gis = GIS("home")
            >>> nb_server = gis.notebook_server[0]

            >>> notebook_item = gis.content.get('<notebook_item_id>')

            >>> nb_mgr = nb_server.notebooks
            >>> nb_mgr.execute_notebook(notebook_item)

        """
        from arcgis.gis import Item

        url = self._url + "/executeNotebook"
        itemid = None
        if isinstance(item, str):
            itemid = item
        elif isinstance(item, Item):
            itemid = item.itemid
        params = {
            "f": "json",
            "itemId": itemid,
            "updatePortalItem": update_portal_item,
            "saveInjectedParameters": save_parameters,
        }
        if parameters:
            params["notebookParameters"] = parameters
        if future:

            def _fn(url, params, nbs):
                import time

                start_job = self._gis._con.post(url, params)
                if "jobUrl" in start_job:
                    resp = self._gis._con.get(start_job["jobUrl"], {"f": "json"})
                else:
                    return start_job
                if "status" in resp and resp["status"].lower() != "success":
                    status = self._gis._con.get(start_job["jobUrl"], {"f": "json"})
                    i = 0
                    while status["status"].lower() != "completed":
                        time.sleep(0.3 * i)
                        if status["status"].lower() == "failed":
                            return status
                        elif (
                            status["status"].lower().find("fail") > -1
                            or status["status"].lower().find("error") > -1
                        ):
                            raise Exception(f"Job Fail {status}")
                        status = self._gis._con.get(start_job["jobUrl"], {"f": "json"})
                        i += 1
                        if i > 20:
                            i = 20
                    return status
                return resp

            return NotebookManager._future_job(
                fn=_fn,
                task_name="Execute Notebook",
                gis=self._gis,
                **{"url": url, "params": params, "nbs": self._nbs},
            )
        res = self._gis._con.post(url, params)
        return res

    # ----------------------------------------------------------------------
    def open_notebook(
        self,
        itemid: str,
        templateid: Optional[str] = None,
        nb_runtimeid: Optional[str] = None,
        template_nb: Optional[str] = None,
        *,
        future: bool = False,
    ):
        """

        Opens a notebook on the notebook server

        ==================      ====================================================================
        **Parameter**            **Description**
        ------------------      --------------------------------------------------------------------
        itemid                  Required String. Opens an existing portal item.
        ------------------      --------------------------------------------------------------------
        templateid              Optional String. The id of the portal notebook template. To get the
                                system templates, look at the sample notebooks group:

                                .. code-block:: python

                                    >>> from arcgis.gis import GIS
                                    >>> gis = GIS()
                                    >>> grp = gis.groups.search("title:(esri sample notebooks) AND
                                    >>>                                 owner:\"esri_notebook\")[0]
                                    >>> grp.content
        ------------------      --------------------------------------------------------------------
        nb_runtimeid            Optional String. The runtime to use to generate a new notebook.
        ------------------      --------------------------------------------------------------------
        template_nb             Optional String. The start up template for the notebook.
        ------------------      --------------------------------------------------------------------
        future                  Optional Bool. If True, the job will run asynchronously.
        ==================      ====================================================================

        :return: Dict or Job

        """

        def _fn(url, params, nbs):
            """used to fire off async job"""
            import time

            start_job = self._gis._con.post(url, params)
            status_url = start_job.get("jobUrl", None) or start_job.get(
                "notebookStatusUrl", None
            )
            if status_url:
                resp = self._gis._con.get(status_url, {"f": "json"})
            else:
                return start_job
            if "status" in resp and resp["status"].lower() != "success":
                status = self._gis._con.get(status_url, {"f": "json"})
                i = 0
                while status["status"].lower() != "completed":
                    time.sleep(0.3 * i)
                    if status["status"].lower() == "failed":
                        return status
                    elif (
                        status["status"].lower().find("fail") > -1
                        or status["status"].lower().find("error") > -1
                    ):
                        raise Exception(f"Job Fail {status}")
                    status = self._gis._con.get(status_url, {"f": "json"})
                    i += 1
                    if i > 20:
                        i = 20
                return status

        params = {
            "itemId": itemid,
            "templateId": templateid,
            "notebookRuntimeId": nb_runtimeid,
            "templateNotebook": template_nb,
            "async": True,
            "f": "json",
        }
        url = self._url + "/openNotebook"
        if future:
            return NotebookManager._future_job(
                fn=_fn,
                task_name="Open Notebook",
                gis=self._gis,
                **{"url": url, "params": params, "nbs": self._nbs},
            )
        else:
            res = self._con.post(url, params)
            status_url = res.get("jobUrl", None) or res.get("notebookStatusUrl", None)
            if status_url:
                job_url = status_url
                params = {"f": "json"}
                job_res = self._con.get(job_url, params)
                while job_res["status"] != "COMPLETED":
                    job_res = self._con.get(job_url, params)
                    if job_res["status"].lower().find("fail") > -1:
                        return job_res
                return job_res
            return res

    # ----------------------------------------------------------------------
    def _add_runtime(
        self,
        name,
        image_id,
        version="10.7",
        container_type="docker",
        image_pull_string="",
        max_cpu=1.0,
        max_memory=4.0,
        max_memory_unit="g",
        max_swap_memory="",
        max_swap_unit="g",
        shared_memory=None,
        shared_memory_unit="m",
        docker_runtime="",
        manifest=None,
        **kwargs,
    ):
        """
        **WARNING: private method, this will change in future releases**

        Added a new docker image to the notebook server.
        """
        url = self._url + "/runtimes/register"
        params = {
            "f": "json",
            "name": name,
            "version": version,
            "imageId": image_id,
            "containerType": container_type,
            "imagePullString": image_pull_string,
            "maxCpu": float(max_cpu),
            "maxMemory": float(max_memory),
            "maxMemoryUnit": max_memory_unit,
            "maxSwapMemory": max_swap_memory,
            "maxSwapMemoryUnit": max_swap_unit,
            "sharedMemory": shared_memory,
            "sharedMemoryUnit": shared_memory_unit,
            "dockerRuntime": docker_runtime,
            "f": "json",
        }

        for k, v in kwargs.items():
            params[k] = v
        res = self._con.post(
            url,
            params,
            files={"manifestFile": manifest},
        )
        return res


########################################################################
class Runtime(object):
    """
    Provides information about the properties of a specific notebook runtime in your ArcGIS Notebook Server site
    """

    _url = None
    _gis = None
    _properties = None

    # ----------------------------------------------------------------------
    def __init__(self, url, gis):
        """Constructor"""
        self._url = url
        if isinstance(gis, GIS):
            self._gis = gis
            self._con = self._gis._con
        else:
            raise ValueError("Invalid GIS object")

    # ----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        try:
            params = {"f": "json"}
            res = self._gis._con.get(self._url, params)
            self._properties = PropertyMap(res)
        except:
            self._properties = PropertyMap({})

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< Runtime @ {url} >".format(url=self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< Runtime @ {url} >".format(url=self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties of the resource"""
        if self._properties is None:
            self._init()
        return self._properties

    # ----------------------------------------------------------------------
    def delete(self):
        """
        Deletes the current runtime from the ArcGIS Notebook Server

        :return: Boolean

        """
        url = self._url + "/unregister"
        params = {"f": "json"}
        res = self._con.post(url, params)
        if "status" in res:
            return res["status"] == "success"
        return res

    # ----------------------------------------------------------------------
    def update(
        self,
        name: Optional[str] = None,
        image_id: Optional[str] = None,
        max_cpu: Optional[float] = None,
        max_memory: Optional[float] = None,
        memory_unit: Optional[str] = None,
        max_swap_memory: Optional[str] = None,
        swap_memory_unit: Optional[str] = None,
        shared_memory: Optional[str] = None,
        docker_runtime: Optional[str] = None,
        shared_unit: Optional[str] = None,
        version: Optional[str] = None,
        container_type: Optional[str] = None,
        pull_string: Optional[str] = None,
        require_advanced_priv: Optional[bool] = None,
        manifest: Optional[str] = None,
    ):
        """
        This operation allows you to update the properties of a notebook
        runtime in ArcGIS Notebook Server. These settings will be applied
        to every container to which the runtime is applied.

        You can use this operation to update the resource limits of the
        runtime, such as maximum CPU and maximum memory. You can also use
        it to extend either of the default notebook runtimes, in order to
        make additional Python modules available to your notebook authors,
        or as a step in making ArcGIS Notebook Server able to use graphical
        processing units (GPUs).



        """
        url = self._url + "/update"
        if manifest is None:
            manifest = ""
        if manifest:
            file = {"manifestFile": manifest}

        params = {
            "name": name,
            "version": version,
            "imageId": image_id,
            "containerType": container_type,
            "imagePullString": pull_string,
            "requiresAdvancedPrivileges": require_advanced_priv,
            "maxCpu": max_cpu or float(self.properties.maxCpu),
            "maxMemory": max_memory or float(self.properties.maxMemory),
            "maxMemoryUnit": memory_unit or "g",
            "maxSwapMemory": max_swap_memory or "",
            "maxSwapMemoryUnit": swap_memory_unit or "g",
            "sharedMemory": shared_memory or "",
            "sharedMemoryUnit": shared_unit or "m",
            "dockerRuntime": docker_runtime,
            "f": "json",
        }
        import json

        for k in list(params.keys()):
            if params[k] is None and k in self.properties:
                params[k] = self.properties[k]
            elif params[k] is None:
                params[k] = ""
            if isinstance(params[k], bool):
                params[k] = json.dumps(params[k])
            elif isinstance(params[k], (int, float)):
                params[k] = float(params[k])

        if len(params) == 1:
            return False
        res = self._con.post(
            url,
            params,
            files={"manifestFile": manifest},
        )
        if "status" in res:
            return res["status"] == "success"
        return res

    # ----------------------------------------------------------------------
    @property
    def manifest(self):
        """
        This resource returns a JSON representation of all the Python
        libraries supported in the specified notebook runtime. Notebook
        authors who open notebooks using this runtime are able to import
        any of the libraries in the manifest into their notebooks.

        :return: List of Dictionaries

        """
        url = self._url + "/manifest"
        params = {"f": "json"}
        res = self._con.get(url, params)
        if "libraries" in res:
            return res["libraries"]
        return res


###########################################################################
class Notebook(object):
    """
    This represents an individual notebook resource in the notebook server.
    """

    _url = None
    _item_id = None
    _properties = None
    _gis = None

    # ----------------------------------------------------------------------
    def __init__(self, url, item_id, properties=None, gis=None):
        self._url = url + "/%s" % item_id
        self._item_id = item_id
        if properties:
            self._properties = properties
        if gis is None:
            from arcgis.env import active_gis

            gis = active_gis
        self._gis = gis

    # ----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        try:
            params = {"f": "json"}
            res = self._gis._con.get(self._url, params)
            self._properties = PropertyMap(res)
        except:
            self._properties = PropertyMap({})

    # ----------------------------------------------------------------------
    def __str__(self):
        return "< Notebook @ {url} >".format(url=self._url)

    # ----------------------------------------------------------------------
    def __repr__(self):
        return "< Notebook @ {url} >".format(url=self._url)

    # ----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties of the resource"""
        if self._properties is None:
            self._init()
        return self._properties

    # ----------------------------------------------------------------------
    def close(self):
        """
        This operation stops a running notebook. You can use it to free up
        space in your notebook container. Idle notebooks are automatically
        cleaned up according to the duration defined by the
        idleNotebookThreshold property. The default value for that property
        is 24 hours.

        :return: Boolean

        """
        params = {"f": "json"}
        url = self._url + "/closeNotebook"
        res = self._gis._con.post(url, params)
        if "status" in res:
            return res["status"] == "success"
        return res



import traceback
from .._utils.env import raise_fastai_import_error

import_exception = None
try:
    from .._data import _raise_fastai_import_error
    from ._arcgis_model import ArcGISModel, SaveModelCallback, _set_multigpu_callback
    from ._pointcnn_utils import (
        PointCNNSeg,
        SamplePointsCallback,
        CrossEntropyPC,
        accuracy,
        CalculateClassificationReport,
        accuracy_non_zero,
        AverageMetric,
        precision,
        recall,
        f1,
    )
    from .._utils.pointcloud_data import (
        get_device,
        inference_las,
        show_results,
        compute_precision_recall,
        predict_h5,
        show_results_tool,
    )
    from ._unet_utils import is_no_color
    from fastai.basic_train import Learner
    import torch
    import numpy as np
    from fastai.callbacks import EarlyStoppingCallback
    from functools import partial
    from ._arcgis_model import _EmptyData
    import json
    from pathlib import Path
    from .._utils.common import _get_emd_path

    HAS_FASTAI = True
except Exception as e:
    import_exception = traceback.format_exc()
    HAS_FASTAI = False


class PointCNN(ArcGISModel):

    """
    Model architecture from https://arxiv.org/abs/1801.07791.
    Creates a Point Cloud classification model.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            :meth:`~arcgis.learn.prepare_data` function.
    ---------------------   -------------------------------------------
    pretrained_path         Optional String. Path where pre-trained model
                            is saved.
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    encoder_params          Optional dictionary. The keys of the dictionary are
                            `out_channels`, `P`, `K`, `D` and `m`.

                            Examples:

                                |    {'out_channels':[16, 32, 64, 96],
                                |    'P':[-1, 768, 384, 128],
                                |    'K':[12, 16, 16, 16],
                                |    'D':[1, 1, 2, 2],
                                |    'm':8
                                |    }

                            Length of `out_channels`, `P`, `K`, `D` should be same.
                            The length denotes the number of layers in encoder.

                            Parameter Explanation

                            - 'out_channels': Number of channels produced by each layer,
                            - 'P': Number of points in each layer,
                            - 'K': Number of K-nearest neighbor in each layer,
                            - 'D': Dilation in each layer,
                            - 'm': Multiplier which is multiplied by each element of out_channel.
    ---------------------   -------------------------------------------
    dropout                 Optional float. This parameter will control overfitting.
                            The range of this parameter is [0,1).
    ---------------------   -------------------------------------------
    sample_point_num        Optional integer. The number of points that the model
                            will actually process.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.PointCNN`  Object
    """

    def __init__(self, data, pretrained_path=None, *args, **kwargs):
        super().__init__(data, None)

        if not HAS_FASTAI:
            raise_fastai_import_error(
                import_exception=import_exception,
                message="This model requires module 'torch_geometric' to be installed.",
                installation_steps=" ",
            )

        self._backbone = None
        self.sample_point_num = kwargs.get("sample_point_num", data.max_point)
        self.learn = Learner(
            data,
            PointCNNSeg(
                self.sample_point_num,
                data.c,
                data.extra_dim,
                kwargs.get("encoder_params", None),
                kwargs.get("dropout", None),
            ),
            loss_func=CrossEntropyPC(data.c),
            metrics=[
                AverageMetric(accuracy),
                AverageMetric(precision),
                AverageMetric(recall),
                AverageMetric(f1),
            ],
            callback_fns=[
                partial(SamplePointsCallback, sample_point_num=self.sample_point_num),
                CalculateClassificationReport,
            ],
        )
        self.encoder_params = self.learn.model.encoder_params

        self.learn.model = self.learn.model.to(self._device)

        if pretrained_path is not None:
            self.load(pretrained_path)

    @staticmethod
    def _available_metrics():
        return ["valid_loss", "accuracy", "precision", "recall", "f1"]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates an PointCNN model object from a Deep Learning Package(DLPK)
        or Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data` function or None for
                                inferencing.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.PointCNN`  Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])
        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file
        model_params = emd["ModelParameters"]

        try:
            class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}
            color_mapping = {i["Value"]: i["Color"] for i in emd["Classes"]}
        except KeyError:
            class_mapping = {i["ClassValue"]: i["ClassName"] for i in emd["Classes"]}
            color_mapping = {i["ClassValue"]: i["Color"] for i in emd["Classes"]}

        if data is None:
            data = _EmptyData(
                path=emd_path.parent.parent,
                loss_func=None,
                c=len(class_mapping),
                chip_size=emd["ImageHeight"],
            )
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd
            for key, value in emd["DataAttributes"].items():
                setattr(data, key, value)

            ## backward compatibility.
            if not hasattr(data, "class_mapping"):
                data.class_mapping = class_mapping
                if not hasattr(data, "class2idx"):
                    data.class2idx = class_mapping
            if not hasattr(data, "color_mapping"):
                data.color_mapping = color_mapping
            if not hasattr(data, "class2idx"):
                data.class2idx = data.class_mapping
            if not hasattr(data, "classes"):
                data.classes = list(data.class2idx.values())

            data.class2idx = {int(k): int(v) for k, v in data.class2idx.items()}
            if hasattr(data, "idx2class"):
                data.idx2class = {int(k): int(v) for k, v in data.idx2class.items()}

            data.color_mapping = {int(k): v for k, v in data.color_mapping.items()}

            ## Below are the lines to make save function work
            data.chip_size = None
            data._image_space_used = None
            data.dataset_type = "PointCloud"

        return cls(data, **model_params, pretrained_path=str(model_file))

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    def fit(
        self,
        epochs=10,
        lr=None,
        one_cycle=True,
        early_stopping=False,
        checkpoint=True,
        tensorboard=False,
        **kwargs,
    ):
        """
        Train the model for the specified number of epochs and using the
        specified learning rates. The precision, recall and f1 scores
        shown in the training table are macro averaged over all classes.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        epochs                  Required integer. Number of cycles of training
                                on the data. Increase it if underfitting.
        ---------------------   -------------------------------------------
        lr                      Optional float or slice of floats. Learning rate
                                to be used for training the model. If ``lr=None``,
                                an optimal learning rate is automatically deduced
                                for training the model.
        ---------------------   -------------------------------------------
        one_cycle               Optional boolean. Parameter to select 1cycle
                                learning rate schedule. If set to `False` no
                                learning rate schedule is used.
        ---------------------   -------------------------------------------
        early_stopping          Optional boolean. Parameter to add early stopping.
                                If set to 'True' training will stop if parameter
                                `monitor` value stops improving for 5 epochs.
                                A minimum difference of 0.001 is required for
                                it to be considered an improvement.
        ---------------------   -------------------------------------------
        checkpoint              Optional boolean or string.
                                Parameter to save checkpoint during training.
                                If set to `True` the best model
                                based on `monitor` will be saved during
                                training. If set to 'all', all checkpoints
                                are saved. If set to False, checkpointing will
                                be off. Setting this parameter loads the best
                                model at the end of training.
        ---------------------   -------------------------------------------
        tensorboard             Optional boolean. Parameter to write the training log.
                                If set to 'True' the log will be saved at
                                <dataset-path>/training_log which can be visualized in
                                tensorboard. Required tensorboardx version=2.1

                                The default value is 'False'.

                                .. note::
                                        Not applicable for Text Models
        ---------------------   -------------------------------------------
        monitor                 Optional string. Parameter specifies
                                which metric to monitor while checkpointing
                                and early stopping. Defaults to 'valid_loss'. Value
                                should be one of the metric that is displayed in
                                the training table. Use `{model_name}.available_metrics`
                                to list the available metrics to set here.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        iters_per_epoch         Optional integer. The number of iterations
                                to run during the training phase.
        =====================   ===========================================

        """
        iterations = kwargs.get("iters_per_epoch", None)
        from ._pointcnn_utils import IterationStop

        callbacks = kwargs["callbacks"] if "callbacks" in kwargs.keys() else []
        if iterations is not None:
            del kwargs["iters_per_epoch"]
            stop_iteration_cb = IterationStop(self.learn, iterations)
            callbacks.append(stop_iteration_cb)
            kwargs["callbacks"] = callbacks
        self._check_requisites()

        if lr is None:
            print("Finding optimum learning rate.")
            lr = self.lr_find(allow_plot=False)

        if isinstance(lr, slice):
            lr = lr.stop

        super().fit(
            epochs, lr, one_cycle, early_stopping, checkpoint, tensorboard, **kwargs
        )

    @property
    def _model_metrics(self):
        return {"accuracy": self._get_model_metrics()}

    def _get_model_metrics(self, **kwargs):
        checkpoint = getattr(self, "_is_checkpointed", False)
        if not hasattr(self.learn, "recorder"):
            return 0.0

        if len(self.learn.recorder.metrics) == 0:
            return 0.0

        model_accuracy = self.learn.recorder.metrics[-1][0]
        if checkpoint:
            val_losses = self.learn.recorder.val_losses
            model_accuracy = self.learn.recorder.metrics[
                self.learn._best_epoch  # index using best epoch.
            ][0]

        return float(model_accuracy)

    def unfreeze(self):
        """
        Not implemented for this model as
        none of the layers are frozen by default.
        """

    def _get_emd_params(self, save_inference_file):
        import random

        _emd_template = {"DataAttributes": {}, "ModelParameters": {}}
        _emd_template["ModelType"] = "PointCloudClassification"
        _emd_template["Framework"] = "N/A"
        _emd_template["ModelConfiguration"] = "N/A"
        _emd_template["ExtractBands"] = "N/A"
        _emd_template["ModelParameters"]["encoder_params"] = self.encoder_params
        _emd_template["ModelParameters"]["sample_point_num"] = self.sample_point_num

        _emd_template["DataAttributes"]["block_size"] = self._data.block_size
        _emd_template["DataAttributes"]["max_point"] = self._data.max_point
        _emd_template["DataAttributes"]["extra_features"] = self._data.extra_features
        _emd_template["DataAttributes"]["extra_dim"] = self._data.extra_dim
        _emd_template["DataAttributes"]["class2idx"] = self._data.class2idx
        _emd_template["DataAttributes"]["color_mapping"] = self._data.color_mapping
        _emd_template["DataAttributes"]["remap"] = self._data.remap
        _emd_template["DataAttributes"]["classes"] = self._data.classes
        _emd_template["DataAttributes"]["subset_classes"] = self._data.subset_classes
        _emd_template["DataAttributes"]["class_mapping"] = self._data.class_mapping
        _emd_template["DataAttributes"]["remap_dict"] = self._data.remap_dict
        _emd_template["DataAttributes"]["remap_bool"] = self._data.remap_bool
        _emd_template["DataAttributes"][
            "important_classes"
        ] = self._data.important_classes
        _emd_template["DataAttributes"]["idx2class"] = self._data.idx2class
        _emd_template["DataAttributes"][
            "features_to_keep"
        ] = self._data.features_to_keep
        _emd_template["DataAttributes"]["pc_type"] = self._data.pc_type
        _emd_template["DataAttributes"][
            "background_classcode"
        ] = self._data.background_classcode

        if hasattr(self.learn.data, "statistics") and self.learn.data.statistics[
            "parameters"
        ].get("excludedClasses", False):
            _emd_template["excludedClasses"] = self.learn.data.statistics["parameters"][
                "excludedClasses"
            ]

        if self._data.pc_type == "PointCloud_TF":
            _emd_template["DataAttributes"][
                "extra_feat_indexes"
            ] = self._data.extra_feat_indexes

        _emd_template["Classes"] = []
        class_data = {}
        for i, class_name in enumerate(self._data.classes):
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = class_name
            class_data["Name"] = self._data.class_mapping[class_name]
            color = (
                [random.choice(range(256)) for i in range(3)]
                if is_no_color(self._data.color_mapping)
                else self._data.color_mapping[class_name]
            )
            class_data["Color"] = np.array(color).astype(int).tolist()
            _emd_template["Classes"].append(class_data.copy())

        if hasattr(self.learn.data, "statistics") and self.learn.data.statistics.get(
            "blockShape", False
        ):
            _emd_template["blockShape"] = self.learn.data.statistics.get("blockShape")

        return _emd_template

    def show_results(self, rows=2, **kwargs):
        """
        Displays the results from your model on the validation set
        with ground truth on the left and predictions on the right.
        Visualization of data, exported in a geographic coordinate system
        is not yet supported.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional rows. Number of rows to show. Default
                                value is 2 and maximum value is the `batch_size`
                                passed in :meth:`~arcgis.learn.prepare_data`.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        color_mapping           Optional dictionary. Mapping from class value
                                to RGB values. Default value example:
                                {0:[220,220,220], 2:[255,0,0], 6:[0,255,0]}.
        ---------------------   -------------------------------------------
        mask_class              Optional list of integers. Array containing
                                class values to mask. Use this parameter to
                                display the classes of interest.
                                Default value is [].
                                Example: All the classes are in [0, 1, 2]
                                to display only class `0` set the mask class
                                parameter to be [1, 2]. List of all classes
                                can be accessed from `data.classes` attribute
                                where `data` is the `Databunch` object returned
                                by :meth:`~arcgis.learn.prepare_data` function.
        ---------------------   -------------------------------------------
        width                   Optional integer. Width of the plot. Default
                                value is 750.
        ---------------------   -------------------------------------------
        height                  Optional integer. Height of the plot. Default
                                value is 512.
        ---------------------   -------------------------------------------
        max_display_point       Optional integer. Maximum number of points
                                to display. Default is 20000. A warning will
                                be raised if the total points to display exceeds
                                this parameter. Setting this parameter will
                                randomly sample the specified number of points
                                and once set, it will be used for future uses.
        =====================   ===========================================
        """

        if self._data.pc_type == "PointCloud_TF":
            return show_results(self, rows, **kwargs)
        elif self._data.pc_type == "PointCloud":
            return show_results_tool(self, rows, **kwargs)

    def predict_las(self, path, output_path=None, print_metrics=False, **kwargs):
        """
        Note: This method has been deprecated starting from `ArcGIS API for
        Python` version 1.9.0.
        Use `Classify Points Using Trained Model` tool  available in 3D Analyst
        extension from ArcGIS Pro 2.8 onwards.
        """

        return inference_las(path, self, output_path, print_metrics, **kwargs)

    def compute_precision_recall(self):
        """
        Computes precision, recall and f1-score on the validation sets.
        """

        return compute_precision_recall(self)

    def predict_h5(self, path, output_path=None, **kwargs):
        """
        This method is used for infrencing using HDF file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        path                    Required string. The path to folder where the HDF
                                files which needs to be predicted are present.
        ---------------------   -------------------------------------------
        output_path             Optional string. The path to folder where to dump
                                the resulting HDF files. Defaults to `results`
                                folder in input path.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        batch_size              Optional integer. The number of blocks to process
                                in one batch. Default is set to 1.
        =====================   ===========================================

        :return: Path where files are dumped.
        """

        return predict_h5(self, path, output_path, **kwargs)

    def _get_name_path(self, name_or_path):
        ispath_like = "\\" in name_or_path or "/" in name_or_path
        if ispath_like:
            path = Path(name_or_path)
            # to make fastai from both path and with name
            if path.is_file():
                name = path.stem
            else:
                name = path.parts[-1]
        else:
            name = name_or_path

        return name, ispath_like

    def load(self, name_or_path):
        """
        Loads a compatible saved model for inferencing or fine tuning from the disk.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Name or Path to
                                Deep Learning Package (DLPK) or
                                Esri Model Definition(EMD) file.
        =====================   ===========================================
        """
        name, ispath_like = self._get_name_path(name_or_path)

        if not Path(name_or_path).is_absolute() and ispath_like:
            name_or_path = str(Path(name_or_path).absolute())

        if Path(name_or_path).is_absolute() and name_or_path.endswith(".emd"):
            emd_path = Path(name_or_path)
        elif Path(name_or_path).is_absolute() and name_or_path.endswith(".pth"):
            emd_path = Path(name_or_path).with_suffix(".emd")
        else:
            emd_path = self.learn.path / "models" / name / f"{name}.emd"

        # from_model
        if not emd_path.exists():
            super().load(name_or_path)
            return
        with open(emd_path) as f:
            emd = json.load(f)

        try:
            class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}
            color_mapping = {i["Value"]: i["Color"] for i in emd["Classes"]}
        except KeyError:
            class_mapping = {i["ClassValue"]: i["ClassName"] for i in emd["Classes"]}
            color_mapping = {i["ClassValue"]: i["Color"] for i in emd["Classes"]}

        dummy_data = _EmptyData(
            path=emd_path.parent,
            loss_func=None,
            c=len(class_mapping),
            chip_size=emd["ImageHeight"],
        )
        dummy_data.emd_path = emd_path
        dummy_data.emd = emd
        for key, value in emd["DataAttributes"].items():
            setattr(dummy_data, key, value)

        # old models
        if not hasattr(dummy_data, "classes"):
            dummy_data.classes = [k for k in list(class_mapping.keys())]

        if not hasattr(dummy_data, "pc_type"):
            dummy_data.pc_type = "PointCloud_TF"

        if not hasattr(self._data, "pc_type"):
            self._data.pc_type = "PointCloud_TF"

        string_mapped_features = {
            "numberOfReturns": "num_returns",
            "returnNumber": "return_num",
            "nearInfrared": "nir",
        }
        inverse_string_mapped_features = {
            v: k for k, v in string_mapped_features.items()
        }

        if not hasattr(dummy_data, "features_to_keep"):
            dummy_data.features_to_keep = [
                inverse_string_mapped_features.get(f[0], f[0])
                for f in dummy_data.extra_features
            ]

        if not hasattr(self._data, "features_to_keep"):
            self._data.features_to_keep = [
                inverse_string_mapped_features.get(f[0], f[0])
                for f in self._data.extra_features
            ]

        # for message
        api_fn_name = "`export_point_dataset` function"
        tool_name = "`Prepare Point Cloud Training Data` tool"
        exported_by = (
            api_fn_name if dummy_data.pc_type == "PointCloud_TF" else tool_name
        )
        training_on = (
            api_fn_name if self._data.pc_type == "PointCloud_TF" else tool_name
        )

        if dummy_data.pc_type != self._data.pc_type:
            raise Exception(
                "Models trained on one type of exported data cannot be trained on other. "
                f"Model was trained on exported data from {exported_by}. "
                f"Usage with data exported by {training_on} is not supported. "
                f"Export the data with {exported_by} for fine-tuning."
            )

        if dummy_data.max_point != self._data.max_point:
            raise Exception(
                "Max points do not match. Please export the data again "
                f"and set max points to same as loaded model i.e. {dummy_data.max_point}."
            )

        if dummy_data.features_to_keep != self._data.features_to_keep:
            if "xyz" in dummy_data.features_to_keep:
                dummy_data.features_to_keep.remove("xyz")
            raise Exception(
                f"Extra features of your data and the model to be loaded do not match. "
                f"Set `extra_features` attribute in `prepare_data` to be {dummy_data.features_to_keep}"
            )

        if dummy_data.classes != self._data.classes:
            raise Exception(
                f"Classes in your data and loaded model do not match. Got {self._data.classes}, required {dummy_data.classes}"
            )

        super().load(name_or_path)
